#!/usr/bin/env python
"""
fitting GUI for XRF display
"""
import time
import copy
from functools import partial
from collections import OrderedDict

from threading import Thread

import json
import numpy as np
import wx
import wx.lib.agw.pycollapsiblepane as CP
import wx.lib.scrolledpanel as scrolled
import wx.dataview as dv
DVSTYLE = dv.DV_SINGLE|dv.DV_VERT_RULES|dv.DV_ROW_LINES

from peakutils import peak

from lmfit import Parameter, Minimizer
from lmfit.printfuncs import gformat

from wxutils import (SimpleText, FloatCtrl, FloatSpin, Choice, Font, pack,
                     Button, Check, HLine, GridPanel, RowPanel, CEN, LEFT,
                     RIGHT, FileSave, GUIColors, FRAMESTYLE, BitmapButton,
                     SetTip, GridPanel, Popup, FloatSpinWithPin, get_icon,
                     fix_filename)

from . import FONTSIZE
from xraydb import (material_mu, xray_edge, materials, add_material,
                    atomic_number, atomic_symbol, xray_line)
from .notebooks import flatnotebook
from .parameter import ParameterPanel
from .periodictable import PeriodicTablePanel

from larch import Group

from ..xrf import xrf_background, MCA, FanoFactors
from ..utils.jsonutils import encode4js, decode4js

from .xrfdisplay_utils import (XRFGROUP, mcaname, XRFRESULTS_GROUP,
                               MAKE_XRFRESULTS_GROUP)

def read_filterdata(flist, _larch):
    """ read filters data"""
    materials = _larch.symtable.get_symbol('_xray._materials')
    out = OrderedDict()
    out['None'] = ('', 0)
    for name in flist:
        if name in materials:
            out[name]  = materials[name]
    return out

def VarChoice(p, default=0, size=(75, -1)):
    return Choice(p, choices=['Fix', 'Vary'],
                  size=size, default=default)

NFILTERS = 4
MIN_CORREL = 0.10

tooltips = {'ptable': 'Select Elements to include in model',
            'step': 'size of step extending to low energy side of peak, fraction of peak height',
            'gamma': 'gamma (lorentzian-like weight) of Voigt function',
            'tail': 'intensity of tail function at low energy side of peak',
            'beta': 'width of tail function at low energy side of peak',
            'sigmax': 'scale sigma from Energy/Noise by this amount',
        }

CompositionUnits = ('ng/mm^2', 'wt %', 'ppm')

Detector_Materials = ['Si', 'Ge']
EFano_Text = 'Peak Widths:  sigma = sqrt(E_Fano * Energy + Noise**2) '
Geom_Text = 'Angles in degrees: 90=normal to surface, 0=grazing surface'
Energy_Text = 'All energies in keV'

xrfmod_setup = """## Set up XRF Model
_xrfmodel = xrf_model(xray_energy={en_xray:.2f}, count_time={count_time:.5f},
                      energy_min={en_min:.2f}, energy_max={en_max:.2f})

_xrfmodel.set_detector(thickness={det_thk:.5f}, material='{det_mat:s}',
                cal_offset={cal_offset:.5f}, cal_slope={cal_slope:.5f},
                vary_cal_offset={cal_vary!r}, vary_cal_slope={cal_vary!r},
                peak_step={peak_step:.5f}, vary_peak_step={peak_step_vary:s},
                peak_tail={peak_tail:.5f}, vary_peak_tail={peak_tail_vary:s},
                peak_beta={peak_beta:.5f}, vary_peak_beta={peak_beta_vary:s},
                peak_gamma={peak_gamma:.5f}, vary_peak_gamma={peak_gamma_vary:s},
                noise={det_noise:.5f}, vary_noise={det_noise_vary:s})"""

xrfmod_scattpeak = """_xrfmodel.add_scatter_peak(name='{peakname:s}', center={_cen:.2f},
                amplitude=1e5, step={_step:.5f}, tail={_tail:.5f}, beta={_beta:.5f},
                sigmax={_sigma:.5f},  vary_center={vcen:s}, vary_step={vstep:s},
                vary_tail={vtail:s}, vary_beta={vbeta:s}, vary_sigmax={vsigma:s})"""

xrfmod_fitscript = """
_xrffitresult = _xrfmodel.fit_spectrum({group:s}, energy_min={emin:.2f}, energy_max={emax:.2f})
_xrfresults.insert(0, _xrffitresult)
"""

xrfmod_filter = "_xrfmodel.add_filter('{name:s}', {thick:.5f}, vary_thickness={vary:s})"
xrfmod_matrix = "_xrfmodel.set_matrix('{name:s}', {thick:.5f}, density={density:.5f})"
xrfmod_pileup = "_xrfmodel.add_pileup(scale={scale:.3f}, vary={vary:s})"
xrfmod_escape = "_xrfmodel.add_escape(scale={scale:.3f}, vary={vary:s})"

xrfmod_savejs = "_xrfresults[{nfit:d}].save('{filename:s}')"

xrfmod_elems = """
for atsym in {elemlist:s}:
    _xrfmodel.add_element(atsym)
#endfor
del atsym"""

Filter_Lengths = ['microns', 'mm', 'cm']
Filter_Materials = ['None', 'air', 'nitrogen', 'helium', 'kapton',
                    'beryllium', 'aluminum', 'mylar', 'pmma']

class FitSpectraFrame(wx.Frame):
    """Frame for Spectral Analysis"""

    def __init__(self, parent, size=(700, 825)):
        self.parent = parent
        self._larch = parent.larch
        symtable = self._larch.symtable
        # fetch current spectra from parent
        if not symtable.has_group(XRFRESULTS_GROUP):
            self._larch.eval(MAKE_XRFRESULTS_GROUP)

        self.xrfresults = symtable.get_symbol(XRFRESULTS_GROUP)
        xrfgroup = symtable.get_group(XRFGROUP)
        mcagroup = getattr(xrfgroup, '_mca')
        self.mca = getattr(xrfgroup, mcagroup)
        self.mcagroup = '%s.%s' % (XRFGROUP, mcagroup)

        efactor = 1.0 if max(self.mca.energy) < 250. else 1000.0

        if self.mca.incident_energy is None:
            self.mca.incident_energy = 20.0
        if self.mca.incident_energy > 250:
            self.mca.incident_energy /= 1000.0

        self.nfit = 0
        self.colors = GUIColors()
        wx.Frame.__init__(self, parent, -1, 'Fit XRF Spectra',
                          size=size, style=wx.DEFAULT_FRAME_STYLE)

        self.wids = {}
        self.owids = {}

        pan = GridPanel(self)
        mca_label = getattr(self.mca, 'label', None)
        if mca_label is None:
            mca_label = getattr(self.mca, 'filename', 'mca')

        self.wids['mca_name'] = SimpleText(pan, mca_label, size=(300, -1), style=LEFT)
        self.wids['btn_calc'] = Button(pan, 'Calculate Model', size=(150, -1),
                                       action=self.onShowModel)
        self.wids['btn_fit'] = Button(pan, 'Fit Model', size=(150, -1),
                                       action=self.onFitModel)

        pan.AddText("  XRF Spectrum: ", colour='#880000')
        pan.Add(self.wids['mca_name'], dcol=3)
        pan.Add(self.wids['btn_calc'], newrow=True)
        pan.Add(self.wids['btn_fit'])

        self.panels = {}
        self.panels['Beam & Detector']  = self.beamdet_page
        self.panels['Filters & Matrix'] = self.materials_page
        self.panels['Elements & Peaks'] = self.elempeaks_page
        self.panels['Fit Results']        = self.fitresult_page
        self.panels['Composition'] = self.composition_page

        self.nb = flatnotebook(pan, self.panels,
                               on_change=self.onNBChanged)
        pan.Add((5, 5), newrow=True)
        pan.Add(self.nb, dcol=5, drow=10, newrow=True)
        pan.pack()

        self.Show()
        self.Raise()

    def onNBChanged(self, event=None):
        pagelabel = self.nb._pages.GetPageText(event.GetSelection()).strip()
        if pagelabel.startswith('Composition'):
            self.UpdateCompositionPage()

    def elempeaks_page(self, **kws):
        "elements and peaks parameters"
        mca = self.parent.mca
        wids = self.wids
        p = GridPanel(self)
        self.selected_elems = []
        self.ptable = PeriodicTablePanel(p, multi_select=True, fontsize=10,
                                         tooltip_msg=tooltips['ptable'],
                                         onselect=self.onElemSelect)

        dstep, dtail, dbeta, dgamma = 0.05, 0.10, 0.5, 0.05
        wids['peak_step'] = FloatSpin(p, value=dstep, digits=3, min_val=0,
                                      max_val=1.0, increment=0.01,
                                      tooltip=tooltips['step'])
        wids['peak_gamma'] = FloatSpin(p, value=dgamma, digits=3, min_val=0,
                                       max_val=10.0, increment=0.01,
                                      tooltip=tooltips['gamma'])
        wids['peak_tail'] = FloatSpin(p, value=dtail, digits=3, min_val=0,
                                      max_val=1.0, increment=0.05,
                                      tooltip=tooltips['tail'])

        wids['peak_beta'] = FloatSpin(p, value=dbeta, digits=3, min_val=0,
                                      max_val=10.0, increment=0.01,
                                      tooltip=tooltips['beta'])
        wids['peak_step_vary'] = VarChoice(p, default=0)
        wids['peak_tail_vary'] = VarChoice(p, default=0)
        wids['peak_gamma_vary'] = VarChoice(p, default=0)
        wids['peak_beta_vary'] = VarChoice(p, default=0)


        btn_from_peaks = Button(p, 'Guess Peaks', size=(150, -1),
                                action=self.onElems_GuessPeaks)
        # tooltip='Guess elements from peak locations')
        btn_from_rois = Button(p, 'Use ROIS as Peaks', size=(150, -1),
                               action=self.onElems_FromROIS)
        btn_clear_elems = Button(p, 'Clear All Peaks', size=(150, -1),
                                 action=self.onElems_Clear)
        wx.CallAfter(self.onElems_GuessPeaks)

        p.AddText('Elements to model:', colour='#880000', dcol=2)
        p.Add((2, 2), newrow=True)
        p.Add(self.ptable, dcol=5, drow=5)
        irow = p.irow

        p.Add(btn_from_peaks,   icol=6, dcol=2, irow=irow)
        p.Add(btn_from_rois,    icol=6, dcol=2, irow=irow+1)
        p.Add(btn_clear_elems,  icol=6, dcol=2, irow=irow+2)
        p.irow += 5

        p.Add((2, 2), newrow=True)
        p.AddText('  Step: ')
        p.Add(wids['peak_step'])
        p.Add(wids['peak_step_vary'])

        p.AddText('  Gamma : ')
        p.Add(wids['peak_gamma'])
        p.Add(wids['peak_gamma_vary'])

        p.Add((2, 2), newrow=True)
        p.AddText('  Beta: ')
        p.Add(wids['peak_beta'])
        p.Add(wids['peak_beta_vary'])

        p.AddText('  Tail: ')
        p.Add(wids['peak_tail'])
        p.Add(wids['peak_tail_vary'])

        p.Add((2, 2), newrow=True)
        p.Add(HLine(p, size=(650, 3)), dcol=8)
        p.Add((2, 2), newrow=True)

        #                name, escale, step, sigmax, beta, tail
        scatter_peaks = (('Elastic',  1.00, 0.05, 1.0, 0.5, 0.10),
                         ('Compton1', 0.97, 0.05, 1.5, 2.0, 0.25),
                         ('Compton2', 0.94, 0.05, 2.0, 2.5, 0.25))
        opts = dict(size=(100, -1), min_val=0, digits=4, increment=0.010)
        for name, escale, dstep, dsigma, dbeta, dtail in scatter_peaks:
            en = escale * self.mca.incident_energy
            t = name.lower()
            vary_en = 1 if t.startswith('compton') else 0

            wids['%s_use'%t] = Check(p, label='Include', default=True)
            wids['%s_cen_vary'%t]   = VarChoice(p, default=vary_en)
            wids['%s_step_vary'%t]  = VarChoice(p, default=0)
            wids['%s_beta_vary'%t]  = VarChoice(p, default=0)
            wids['%s_tail_vary'%t]  = VarChoice(p, default=0)
            wids['%s_sigma_vary'%t] = VarChoice(p, default=0)

            wids['%s_cen'%t]  = FloatSpin(p, value=en, digits=3, min_val=0,
                                           increment=0.01)
            wids['%s_step'%t] = FloatSpin(p, value=dstep, digits=3, min_val=0,
                                          max_val=1.0, increment=0.01,
                                          tooltip=tooltips['step'])
            wids['%s_tail'%t] = FloatSpin(p, value=dtail, digits=3, min_val=0,
                                          max_val=1.0, increment=0.05,
                                          tooltip=tooltips['tail'])
            wids['%s_beta'%t] = FloatSpin(p, value=dbeta, digits=3, min_val=0,
                                          max_val=10.0, increment=0.10,
                                          tooltip=tooltips['beta'])
            wids['%s_sigma'%t] = FloatSpin(p, value=dsigma, digits=3, min_val=0,
                                           max_val=10.0, increment=0.05,
                                           tooltip=tooltips['sigmax'])

            p.Add((2, 2), newrow=True)
            p.AddText("  %s Peak:" % name,  colour='#880000')
            p.Add(wids['%s_use' % t], dcol=2)

            p.AddText('  Energy (keV): ')
            p.Add(wids['%s_cen'%t])
            p.Add(wids['%s_cen_vary'%t])

            p.Add((2, 2), newrow=True)
            p.AddText('  Step: ')
            p.Add(wids['%s_step'%t])
            p.Add(wids['%s_step_vary'%t])

            p.AddText('  Sigma Scale : ')
            p.Add(wids['%s_sigma'%t])
            p.Add(wids['%s_sigma_vary'%t])

            p.Add((2, 2), newrow=True)
            p.AddText('  Beta : ')
            p.Add(wids['%s_beta'%t])
            p.Add(wids['%s_beta_vary'%t])

            p.AddText('  Tail: ')
            p.Add(wids['%s_tail'%t])
            p.Add(wids['%s_tail_vary'%t])

            p.Add((2, 2), newrow=True)
            p.Add(HLine(p, size=(650, 3)), dcol=7)

        p.pack()
        return p

    def beamdet_page(self, **kws):
        "beam / detector settings"
        mca = self.mca
        en_min = 2.0
        en_max = self.mca.incident_energy

        cal_offset = getattr(mca, 'offset',  0)
        cal_slope = getattr(mca, 'slope',  0.010)
        det_noise = getattr(mca, 'det_noise',  0.035)
        escape_amp = getattr(mca, 'escape_amp', 1.0)
        pileup_amp = getattr(mca, 'pileup_amp', 0.1)

        wids = self.wids
        # main = wx.Panel(self)
        pdet = GridPanel(self, itemstyle=LEFT)

        def addLine(pan):
            pan.Add(HLine(pan, size=(650, 3)), dcol=6, newrow=True)


        wids['escape_use'] = Check(pdet, label='Include Escape in Fit',
                                   default=True, action=self.onUsePileupEscape)
        wids['escape_amp'] = FloatSpin(pdet, value=escape_amp,
                                         min_val=0, max_val=100, digits=2,
                                         increment=0.02, size=(100, -1))

        wids['pileup_use'] = Check(pdet, label='Include Pileup in Fit',
                                   default=True, action=self.onUsePileupEscape)
        wids['pileup_amp'] = FloatSpin(pdet, value=pileup_amp,
                                         min_val=0, max_val=100, digits=2,
                                         increment=0.02, size=(100, -1))

        wids['escape_amp_vary'] = VarChoice(pdet, default=True)
        wids['pileup_amp_vary'] = VarChoice(pdet, default=True)


        wids['cal_slope'] = FloatSpin(pdet, value=cal_slope,
                                      min_val=0, max_val=100,
                                      digits=4, increment=0.01, size=(100, -1))
        wids['cal_offset'] = FloatSpin(pdet, value=cal_offset,
                                      min_val=-500, max_val=500,
                                      digits=4, increment=0.01, size=(100, -1))

        wids['cal_vary'] = Check(pdet, label='Vary Calibration in Fit', default=True)

        wids['det_mat'] = Choice(pdet, choices=Detector_Materials,
                                 size=(70, -1), default=0,
                                 action=self.onDetMaterial)

        wids['det_thk'] = FloatSpin(pdet, value=0.400, size=(100, -1),
                                     increment=0.010, min_val=0, max_val=10,
                                     digits=4)

        wids['det_noise_vary'] = VarChoice(pdet, default=1)

        opts = dict(size=(100, -1), min_val=0, max_val=500, digits=3,
                    increment=0.10)
        wids['en_xray'] = FloatSpin(pdet, value=self.mca.incident_energy,
                                    action=self.onSetXrayEnergy, **opts)
        wids['en_min'] = FloatSpin(pdet, value=en_min, **opts)
        wids['en_max'] = FloatSpin(pdet, value=en_max, **opts)
        wids['flux_in'] = FloatCtrl(pdet, value=5.e10, gformat=True,
                                    minval=0, size=(100, -1))

        opts.update({'increment': 0.005})
        wids['det_noise'] = FloatSpin(pdet, value=det_noise, **opts)
        wids['det_efano'] = SimpleText(pdet, size=(200, -1),
                                       label='E_Fano= %.4e' % FanoFactors['Si'])

        opts.update(digits=1, max_val=90, min_val=0, increment=1)
        wids['angle_in'] = FloatSpin(pdet, value=45, **opts)
        wids['angle_out'] = FloatSpin(pdet, value=45, **opts)

        opts.update(digits=1, max_val=5e9, min_val=0, increment=1)
        wids['det_dist'] = FloatSpin(pdet, value=50, **opts)
        wids['det_area'] = FloatSpin(pdet, value=50, **opts)

        for notyet in ('angle_in', 'angle_out', 'det_dist', 'det_area',
                       'flux_in'):
            wids[notyet].Disable()

        pdet.AddText(' Beam Energy, Fit Range :', colour='#880000', dcol=2)
        pdet.AddText('   X-ray Energy (keV): ', newrow=True)
        pdet.Add(wids['en_xray'])
        pdet.AddText('Incident Flux (Hz): ', newrow=False)
        pdet.Add(wids['flux_in'])
        pdet.AddText('   Fit Energy Min (keV): ', newrow=True)
        pdet.Add(wids['en_min'])
        pdet.AddText('Fit Energy Max (keV): ')
        pdet.Add(wids['en_max'])


        addLine(pdet)
        pdet.AddText(' Energy Calibration :', colour='#880000', dcol=1, newrow=True)
        pdet.Add(wids['cal_vary'], dcol=2)
        pdet.AddText('   Offset (keV): ', newrow=True)
        pdet.Add(wids['cal_offset'])
        pdet.AddText('Slope (keV/bin): ')
        pdet.Add(wids['cal_slope'])

        addLine(pdet)
        pdet.AddText(' Detector Material:', colour='#880000', dcol=1, newrow=True)
        pdet.AddText(EFano_Text, dcol=3)
        pdet.AddText('   Material:  ', newrow=True)
        pdet.Add(wids['det_mat'])
        pdet.Add(wids['det_efano'], dcol=2)
        pdet.AddText('   Thickness (mm): ', newrow=True)
        pdet.Add(wids['det_thk'])
        pdet.AddText('   Noise (keV): ', newrow=True)
        pdet.Add(wids['det_noise'])
        pdet.Add(wids['det_noise_vary'], dcol=2)


        addLine(pdet)
        pdet.AddText(' Escape && Pileup:', colour='#880000', dcol=2, newrow=True)
        pdet.AddText('   Escape Scale:', newrow=True)
        pdet.Add(wids['escape_amp'])
        pdet.Add(wids['escape_amp_vary'])
        pdet.Add(wids['escape_use'], dcol=3)

        pdet.AddText('   Pileup Scale:', newrow=True)
        pdet.Add(wids['pileup_amp'])
        pdet.Add(wids['pileup_amp_vary'])
        pdet.Add(wids['pileup_use'], dcol=3)

        addLine(pdet)
        pdet.AddText(' Geometry:', colour='#880000', dcol=1, newrow=True)
        pdet.AddText(Geom_Text, dcol=3)
        pdet.AddText('   Incident Angle (deg):', newrow=True)
        pdet.Add(wids['angle_in'])
        pdet.AddText('   Exit Angle (deg):', newrow=False)
        pdet.Add(wids['angle_out'])
        pdet.AddText('   Detector Distance (mm): ', newrow=True)
        pdet.Add(wids['det_dist'])
        pdet.AddText('   Detector Area (mm^2): ', newrow=False)
        pdet.Add(wids['det_area'])


        addLine(pdet)
        pdet.pack()
        return pdet

    def materials_page(self, **kws):
        "filters and matrix settings"
        wids = self.wids
        pan = GridPanel(self, itemstyle=LEFT)

        pan.AddText(' Filters :', colour='#880000', dcol=2) # , newrow=True)
        pan.AddManyText(('  Filter #', 'Material', 'Thickness (mm)',
                         'Vary Thickness'), style=LEFT, newrow=True)
        opts = dict(size=(125, -1), min_val=0, digits=5, increment=0.005)

        for i in range(NFILTERS):
            t = 'filter%d' % (i+1)
            wids['%s_mat'%t] = Choice(pan, choices=Filter_Materials, default=0,
                                      size=(150, -1),
                                      action=partial(self.onFilterMaterial, index=i+1))
            wids['%s_thk'%t] = FloatSpin(pan, value=0.0, **opts)
            wids['%s_var'%t] = VarChoice(pan, default=0)
            if i == 0: # first selection
                wids['%s_mat'%t].SetStringSelection('beryllium')
                wids['%s_thk'%t].SetValue(0.0250)
            elif i == 1: # second selection
                wids['%s_mat'%t].SetStringSelection('air')
                wids['%s_thk'%t].SetValue(50.00)
            elif i == 2: # third selection
                wids['%s_mat'%t].SetStringSelection('kapton')
                wids['%s_thk'%t].SetValue(0.00)
            elif i == 3: # third selection
                wids['%s_mat'%t].SetStringSelection('aluminum')
                wids['%s_thk'%t].SetValue(0.00)

            pan.AddText('      %i' % (i+1), newrow=True)
            pan.Add(wids['%s_mat' % t])
            pan.Add(wids['%s_thk' % t])
            pan.Add(wids['%s_var' % t])

        pan.Add(HLine(pan, size=(650, 3)), dcol=6, newrow=True)

        pan.AddText(' Matrix:', colour='#880000', newrow=True)
        pan.AddText('    NOTE: thin film limit only',  dcol=3)

        wids['matrix_mat'] = wx.TextCtrl(pan, value='', size=(275, -1))
        wids['matrix_thk'] = FloatSpin(pan, value=0.0, **opts)
        wids['matrix_den'] = FloatSpin(pan, value=1.0, **opts)
        wids['matrix_btn'] = Button(pan, 'Use Material', size=(175, -1),
                                  action=self.onUseCurrentMaterialAsFilter)
        wids['matrix_btn'].Disable()
        pan.AddText('  Material/Formula:', dcol=1, newrow=True)
        pan.Add(wids['matrix_mat'], dcol=2)
        pan.Add(wids['matrix_btn'], dcol=3)
        pan.AddText('  Thickness (mm):', newrow=True)
        pan.Add(wids['matrix_thk'])
        pan.AddText(' Density (gr/cm^3):', newrow=False)
        pan.Add(wids['matrix_den'])

        pan.Add(HLine(pan, size=(650, 3)), dcol=6, newrow=True)

        # Materials
        pan.AddText(' Known Materials:', colour='#880000', dcol=4, newrow=True)

        mview = self.owids['materials'] = dv.DataViewListCtrl(pan, style=DVSTYLE)
        mview.Bind(dv.EVT_DATAVIEW_SELECTION_CHANGED, self.onSelectMaterial)
        self.selected_material = ''


        mview.AppendTextColumn('Name',      width=150)
        mview.AppendTextColumn('Formula',   width=325)
        mview.AppendTextColumn('density',    width=90)
        mview.AppendToggleColumn('Filter?',  width=75)
        for col in range(4):
            this = mview.Columns[col]
            align = wx.ALIGN_LEFT
            this.Sortable = True
            this.Alignment = this.Renderer.Alignment = align

        mview.SetMinSize((675, 170))
        mview.DeleteAllItems()
        self.materials_data = {}
        for name, data in materials._read_materials_db().items():
            # print("DATA " , name, data)
            formula, density = data.formula, data.density
            self.materials_data[name] = (formula, density)
            mview.AppendItem((name, formula, "%9.6f"%density,
                              name in Filter_Materials))
        pan.Add(mview, dcol=5, newrow=True)

        pan.AddText(' Add Material:', colour='#880000', newrow=True)
        pan.Add(Button(pan, 'Add', size=(175, -1),
                       action=self.onAddMaterial))
        pan.Add((10, 10))
        bx = Button(pan, 'Update Filter List', size=(175, -1),
                    action=self.onUpdateFilterList)
        pan.Add(bx)

        self.owids['newmat_name'] = wx.TextCtrl(pan, value='', size=(175, -1))
        self.owids['newmat_dens'] = FloatSpin(pan, value=1.0, **opts)
        self.owids['newmat_form'] = wx.TextCtrl(pan, value='', size=(400, -1))


        for notyet in ('matrix_mat', 'matrix_thk', 'matrix_den',
                       'matrix_btn'):
            wids[notyet].Disable()

        pan.AddText(' Name:', newrow=True)
        pan.Add(self.owids['newmat_name'])
        pan.AddText(' Density (gr/cm^3):', newrow=False)
        pan.Add(self.owids['newmat_dens'])
        pan.AddText(' Formula:', newrow=True)
        pan.Add(self.owids['newmat_form'], dcol=3)
        pan.pack()
        return pan

    def fitresult_page(self, **kws):
        sizer = wx.GridBagSizer(10, 5)
        panel = scrolled.ScrolledPanel(self)
        # title row
        wids = self.owids
        title = SimpleText(panel, 'Fit Results', font=Font(FONTSIZE+1),
                           colour=self.colors.title, style=LEFT)

        wids['data_title'] = SimpleText(panel, '< > ', font=Font(FONTSIZE+1),
                                             colour=self.colors.title, style=LEFT)

        wids['fitlabel_lab'] = SimpleText(panel, ' Fit Label: ')
        wids['fitlabel_txt'] = wx.TextCtrl(panel, -1, ' ', size=(150, -1))
        wids['fitlabel_btn'] = Button(panel, 'Set Label',  size=(150, -1),
                                      action=self.onChangeFitLabel)

        opts = dict(default=False, size=(175, -1), action=self.onPlot)
        wids['plot_comps'] = Check(panel, label='Show Components?', **opts)
        self.plot_choice = Button(panel, 'Plot',
                                  size=(150, -1), action=self.onPlot)

        self.save_result = Button(panel, 'Save Model',
                                  size=(150, -1), action=self.onSaveFitResult)
        SetTip(self.save_result, 'save model and result to be loaded later')

        self.export_fit  = Button(panel, 'Export Fit',
                                  size=(150, -1), action=self.onExportFitResult)
        SetTip(self.export_fit, 'save arrays and results to text file')

        irow = 0
        sizer.Add(title,              (irow, 0), (1, 1), LEFT)
        sizer.Add(wids['data_title'], (irow, 1), (1, 3), LEFT)

        irow += 1
        sizer.Add(self.save_result,   (irow, 0), (1, 1), LEFT)
        sizer.Add(self.export_fit,    (irow, 1), (1, 1), LEFT)
        sizer.Add(self.plot_choice,   (irow, 2), (1, 1), LEFT)
        sizer.Add(wids['plot_comps'], (irow, 3), (1, 1), LEFT)

        irow += 1
        sizer.Add(wids['fitlabel_lab'], (irow, 0), (1, 1), LEFT)
        sizer.Add(wids['fitlabel_txt'], (irow, 1), (1, 1), LEFT)
        sizer.Add(wids['fitlabel_btn'], (irow, 2), (1, 2), LEFT)


        irow += 1
        sizer.Add(HLine(panel, size=(650, 3)), (irow, 0), (1, 5), LEFT)

        irow += 1
        title = SimpleText(panel, '[[Fit Statistics]]',  font=Font(FONTSIZE+1),
                           colour=self.colors.title, style=LEFT)
        sizer.Add(title, (irow, 0), (1, 4), LEFT)

        sview = wids['stats'] = dv.DataViewListCtrl(panel, style=DVSTYLE)
        sview.Bind(dv.EVT_DATAVIEW_SELECTION_CHANGED, self.onSelectFit)
        sview.AppendTextColumn('  Fit Label', width=90)
        sview.AppendTextColumn(' N_vary', width=65)
        sview.AppendTextColumn(' N_eval', width=65)
        sview.AppendTextColumn(' \u03c7\u00B2', width=125)
        sview.AppendTextColumn(' \u03c7\u00B2_reduced', width=125)
        sview.AppendTextColumn(' Akaike Info', width=125)

        for col in range(sview.ColumnCount):
            this = sview.Columns[col]
            isort, align = True, wx.ALIGN_RIGHT
            if col == 0:
                align = wx.ALIGN_CENTER
            this.Sortable = isort
            this.Alignment = this.Renderer.Alignment = align
        sview.SetMinSize((675, 150))

        irow += 1
        sizer.Add(sview, (irow, 0), (1, 5), LEFT)

        irow += 1
        sizer.Add(HLine(panel, size=(650, 3)), (irow, 0), (1, 5), LEFT)

        irow += 1
        title = SimpleText(panel, '[[Variables]]',  font=Font(FONTSIZE+1),
                           colour=self.colors.title, style=LEFT)
        sizer.Add(title, (irow, 0), (1, 1), LEFT)

        pview = wids['params'] = dv.DataViewListCtrl(panel, style=DVSTYLE)
        wids['paramsdata'] = []
        pview.AppendTextColumn('Parameter',      width=150)
        pview.AppendTextColumn('Refined Value',  width=100)
        pview.AppendTextColumn('Standard Error', width=100)
        pview.AppendTextColumn('% Uncertainty', width=100)
        pview.AppendTextColumn('Initial Value',  width=150)

        for col in range(4):
            this = pview.Columns[col]
            align = wx.ALIGN_LEFT
            if col > 0:
                align = wx.ALIGN_RIGHT
            this.Sortable = False
            this.Alignment = this.Renderer.Alignment = align

        pview.SetMinSize((675, 200))
        pview.Bind(dv.EVT_DATAVIEW_SELECTION_CHANGED, self.onSelectParameter)

        irow += 1
        sizer.Add(pview, (irow, 0), (1, 5), LEFT)

        irow += 1
        sizer.Add(HLine(panel, size=(650, 3)), (irow, 0), (1, 5), LEFT)

        irow += 1
        title = SimpleText(panel, '[[Correlations]]',  font=Font(FONTSIZE+1),
                           colour=self.colors.title, style=LEFT)

        wids['all_correl'] = Button(panel, 'Show All',
                                    size=(100, -1), action=self.onAllCorrel)

        wids['min_correl'] = FloatSpin(panel, value=MIN_CORREL,
                                       min_val=0, size=(100, -1),
                                       digits=3, increment=0.1)

        ctitle = SimpleText(panel, 'minimum correlation: ')
        sizer.Add(title,  (irow, 0), (1, 1), LEFT)
        sizer.Add(ctitle, (irow, 1), (1, 1), LEFT)
        sizer.Add(wids['min_correl'], (irow, 2), (1, 1), LEFT)
        sizer.Add(wids['all_correl'], (irow, 3), (1, 1), LEFT)

        cview = wids['correl'] = dv.DataViewListCtrl(panel, style=DVSTYLE)

        cview.AppendTextColumn('Parameter 1',    width=150)
        cview.AppendTextColumn('Parameter 2',    width=150)
        cview.AppendTextColumn('Correlation',    width=150)

        for col in (0, 1, 2):
            this = cview.Columns[col]
            this.Sortable = False
            align = wx.ALIGN_LEFT
            if col == 2:
                align = wx.ALIGN_RIGHT
            this.Alignment = this.Renderer.Alignment = align
        cview.SetMinSize((675, 125))

        irow += 1
        sizer.Add(cview, (irow, 0), (1, 5), LEFT)
        pack(panel, sizer)
        panel.SetMinSize((675, 725))
        panel.SetupScrolling()
        return panel

    def composition_page(self, **kws):
        sizer = wx.GridBagSizer(10, 5)
        panel = scrolled.ScrolledPanel(self)
        wids = self.owids
        title = SimpleText(panel, 'Composition Results', font=Font(FONTSIZE+1),
                           colour=self.colors.title, style=LEFT)
        wids['data_title2'] = SimpleText(panel, '< > ', font=Font(FONTSIZE+1),
                                             colour=self.colors.title, style=LEFT)

        cview = wids['composition'] = dv.DataViewListCtrl(panel, style=DVSTYLE)
        cview.AppendTextColumn(' Z ', width=50)
        cview.AppendTextColumn(' Element ', width=100)
        cview.AppendTextColumn(' Amplitude', width=150)
        cview.AppendTextColumn(' Concentration',  width=150)
        cview.AppendTextColumn(' Uncertainty',  width=150)

        for col in range(5):
            this = cview.Columns[col]
            align = wx.ALIGN_RIGHT
            if col ==  1:
                align = wx.ALIGN_LEFT
            this.Sortable = True
            this.Alignment = this.Renderer.Alignment = align

        cview.SetMinSize((675, 500))
        wids['comp_fitlabel'] = Choice(panel, choices=[''], size=(175, -1),
                                       action=self.onCompSelectFit)

        self.compscale_lock = 0.0
        wids['comp_elemchoice'] = Choice(panel, choices=[''], size=(100, -1))
        # action=self.onCompSetElemAbundance)
        wids['comp_elemscale'] = FloatSpin(panel, value=1.0, digits=5, min_val=0,
                                           increment=0.01,
                                           action=self.onCompSetElemAbundance)
        wids['comp_units'] = Choice(panel, choices=CompositionUnits, size=(100, -1))
        wids['comp_scale'] = FloatCtrl(panel, value=0, size=(200, -1), precision=5,
                                       minval=0, action=self.onCompSetScale)

        wids['comp_save'] = Button(panel, 'Save This Concentration Data',
                                   size=(200, -1), action=self.onCompSave)

        irow = 0
        sizer.Add(title,              (irow, 0), (1, 2), LEFT)
        sizer.Add(wids['data_title2'], (irow, 2), (1, 5), LEFT)
        irow += 1
        sizer.Add(SimpleText(panel, 'Fit Label:'),  (irow, 0), (1, 1), LEFT)
        sizer.Add(wids['comp_fitlabel'],            (irow, 1), (1, 5), LEFT)

        irow += 1
        sizer.Add(SimpleText(panel, 'Scale Element:'),   (irow, 0), (1, 1), LEFT)
        sizer.Add(wids['comp_elemchoice'],         (irow, 1), (1, 1), LEFT)
        sizer.Add(SimpleText(panel, ' to:'),       (irow, 2), (1, 1), LEFT)
        sizer.Add(wids['comp_elemscale'],          (irow, 3), (1, 1), LEFT)
        sizer.Add(wids['comp_units'],              (irow, 4), (1, 1), LEFT)

        irow += 1
        sizer.Add(SimpleText(panel, 'Scaling Factor:'), (irow, 0), (1, 1), LEFT)
        sizer.Add(wids['comp_scale'],              (irow, 1), (1, 3), LEFT)

        irow += 1
        sizer.Add(wids['composition'],   (irow, 0), (3, 6), LEFT)

        irow += 3
        sizer.Add(wids['comp_save'],   (irow, 0), (1, 3), LEFT)

        pack(panel, sizer)
        panel.SetMinSize((675, 750))
        panel.SetupScrolling()
        return panel

    def onCompSetScale(self, event=None, value=None):
        if len(self.xrfresults) < 1 or (time.time() - self.compscale_lock) < 0.25:
            return
        self.compscale_lock = time.time()
        owids = self.owids
        result = self.get_fitresult(nfit=owids['comp_fitlabel'].GetSelection())
        cur_elem  = owids['comp_elemchoice'].GetStringSelection()
        conc_vals = {}
        for elem in result.comps.keys():
            parname = 'amp_%s' % elem.lower()
            if parname in result.params:
                par = result.params[parname]
                conc_vals[elem] = [par.value, par.stderr]

        try:
            scale = self.owids['comp_scale'].GetValue()
        except:
            return

        owids['comp_elemscale'].SetValue(conc_vals[cur_elem][0]*scale)
        owids['composition'].DeleteAllItems()
        result.concentration_results = conc_vals
        result.concentration_scale = scale

        for elem, dat in conc_vals.items():
            zat = "%d" % atomic_number(elem)
            val, serr = dat
            rval = "%15.4f" % val
            sval = "%15.4f" % (val*scale)
            uval = "%15.4f" % (serr*scale)
            try:
                uval = uval + ' ({:.2%})'.format(abs(serr/val))
            except ZeroDivisionError:
                pass
            owids['composition'].AppendItem((zat, elem, rval, sval, uval))

    def onCompSetElemAbundance(self, event=None, value=None):
        if len(self.xrfresults) < 1  or (time.time() - self.compscale_lock) < 0.25:
            return
        self.compscale_lock = time.time()
        owids = self.owids
        result = self.get_fitresult(nfit=owids['comp_fitlabel'].GetSelection())
        cur_elem = owids['comp_elemchoice'].GetStringSelection()
        conc_vals = {}
        for elem in result.comps.keys():
            parname = 'amp_%s' % elem.lower()
            if parname in result.params:
                par = result.params[parname]
                conc_vals[elem] = [par.value, par.stderr]

        result.concentration_results = conc_vals
        elem_value = owids['comp_elemscale'].GetValue()

        scale = elem_value/conc_vals[cur_elem][0]
        result.concentration_scale = scale
        owids['comp_scale'].SetValue(scale)
        owids['composition'].DeleteAllItems()
        for elem, dat in conc_vals.items():
            zat = "%d" % atomic_number(elem)
            val, serr = dat
            rval = "%15.4f" % val
            sval = "%15.4f" % (val*scale)
            uval = "%15.4f" % (serr*scale)
            try:
                uval = uval + ' ({:.2%})'.format(abs(serr/val))
            except ZeroDivisionError:
                pass
            owids['composition'].AppendItem((zat, elem, rval, sval, uval))


    def onCompSave(self, event=None):
        result = self.get_fitresult(nfit=self.owids['comp_fitlabel'].GetSelection())
        scale = result.concentration_scale
        deffile = self.mca.label + '_' + result.label
        deffile = fix_filename(deffile.replace('.', '_')) + '_xrf.csv'
        wcards = "CSV (*.csv)|*.csv|All files (*.*)|*.*"
        sfile = FileSave(self, 'Save Concentration Results',
                         default_file=deffile,
                         wildcard=wcards)
        if sfile is not None:
            buff = ["# results for MCA labeled: %s" % self.mca.label,
                    "# fit label: %s" % result.label,
                    "# concentration units: %s" % self.owids['comp_units'].GetStringSelection(),
                    "# count time: %s" % result.count_time,
                    "# scale: %s" % result.concentration_scale,
                    "# Fit Report:"     ]
            for l in result.fit_report.split('\n'):
                buff.append("#    %s" % l)
            buff.append("###########")
            buff.append("#Element  Concentration  Uncertainty  RawAmplitude")
            for elem, dat in result.concentration_results.items():
                eout = (elem  + ' '*4)[:4]
                val, serr = dat
                rval = "%16.5f" % val
                sval = "%16.5f" % (val/scale)
                uval = "%16.5f" % (serr/scale)
                buff.append("  ".join([elem, sval, uval, rval]))
            buff.append('')
            with open(sfile, 'w') as fh:
                fh.write('\n'.join(buff))

    def onCompSelectFit(self, event=None):
        result = self.get_fitresult(nfit=self.owids['comp_fitlabel'].GetSelection())
        cur_elem  = self.owids['comp_elemchoice'].GetStringSelection()
        self.owids['comp_elemchoice'].Clear()
        elems = [el['symbol'] for el in result.elements]
        self.owids['comp_elemchoice'].SetChoices(elems)
        if len(cur_elem) > 0:
            self.owids['comp_elemchoice'].SetStringSelection(cur_elem)
        else:
            self.owids['comp_elemchoice'].SetSelection(0)
        self.onCompSetElemAbundance()

    def UpdateCompositionPage(self, event=None):
        self.xrfresults = self._larch.symtable.get_symbol(XRFRESULTS_GROUP)
        if len(self.xrfresults) > 0:
            result = self.get_fitresult()
            fitlab = self.owids['comp_fitlabel']
            fitlab.Clear()
            fitlab.SetChoices([a.label for a in self.xrfresults])
            fitlab.SetStringSelection(result.label)
            self.onCompSelectFit()

    def onElems_Clear(self, event=None):
        self.ptable.on_clear_all()

    def onElems_GuessPeaks(self, event=None):
        mca = self.mca
        _indices = peak.indexes(mca.counts*1.0, min_dist=5, thres=0.025)
        peak_energies = mca.energy[_indices]

        elrange = range(10, 92)
        atsyms  = [atomic_symbol(i) for i in elrange]
        kalphas = [0.001*xray_line(i, 'Ka').energy for i in elrange]
        kbetas  = [0.001*xray_line(i, 'Kb').energy for i in elrange]
        self.ptable.on_clear_all()
        elems = []
        for iz, en in enumerate(peak_energies):
            for i, ex in enumerate(kalphas):
                if abs(en - ex) < 0.025:
                    elems.append(atsyms[i])
                    peak_energies[iz] = -ex

        for iz, en in enumerate(peak_energies):
            if en > 0:
                for i, ex in enumerate(kbetas):
                    if abs(en - ex) < 0.025:
                        if atsyms[i] not in elems:
                            elems.append(atsyms[i])
                        peak_energies[iz] = -ex

        en = self.wids['en_xray'].GetValue()
        emin = self.wids['en_min'].GetValue()
        for elem in elems:
            kedge  = 0.001*xray_edge(elem, 'K').energy
            l3edge = 0.001*xray_edge(elem, 'L3').energy
            l2edge = 0.001*xray_edge(elem, 'L3').energy
            if ((kedge < en and kedge > emin) or
                (l3edge < en and l3edge > emin) or
                (l2edge < en and l2edge > emin)):
                if elem not in self.ptable.selected:
                    self.ptable.onclick(label=elem)

    def onElems_FromROIS(self, event=None):
        for roi in self.mca.rois:
            words = roi.name.split()
            elem = words[0].title()
            if (elem in self.ptable.syms and
                elem not in self.ptable.selected):
                self.ptable.onclick(label=elem)
        self.onSetXrayEnergy()

    def onSetXrayEnergy(self, event=None):
        en = self.wids['en_xray'].GetValue()
        self.wids['en_max'].SetValue(en)
        self.wids['elastic_cen'].SetValue(en)
        self.wids['compton1_cen'].SetValue(en*0.975)
        self.wids['compton2_cen'].SetValue(en*0.950)
        emin = self.wids['en_min'].GetValue() * 1.25

        self.ptable.on_clear_all()
        for roi in self.mca.rois:
            words = roi.name.split()
            elem = words[0].title()
            kedge = l3edge = l2edge = 0.0
            try:
                kedge  = 0.001*xray_edge(elem, 'K').energy
                l3edge = 0.001*xray_edge(elem, 'L3').energy
                l2edge = 0.001*xray_edge(elem, 'L3').energy
            except:
                pass
            if ((kedge < en and kedge > emin) or
                (l3edge < en and l3edge > emin) or
                (l2edge < en and l2edge > emin)):
                if elem not in self.ptable.selected:
                    self.ptable.onclick(label=elem)

    def onDetMaterial(self, event=None):
        dmat = self.wids['det_mat'].GetStringSelection()
        if dmat not in FanoFactors:
            dmat = 'Si'
        self.wids['det_efano'].SetLabel('E_Fano= %.4e' % FanoFactors[dmat])

    def onFilterMaterial(self, evt=None, index=1):
        name = evt.GetString()
        den = self.materials_data.get(name, (None, 1.0))[1]
        t = 'filter%d' % (index)
        thick = self.wids['%s_thk'%t]
        if den < 0.1 and thick.GetValue() < 0.1:
            thick.SetValue(10.0)
            thick.SetIncrement(0.5)
        elif den > 0.1 and thick.GetValue() < 1.e-5:
            thick.SetValue(0.0250)
            thick.SetIncrement(0.005)

    def onUseCurrentMaterialAsFilter(self, evt=None):
        name = self.selected_material
        density = self.materials_data.get(name, (None, 1.0))[1]
        self.wids['matrix_den'].SetValue(density)
        self.wids['matrix_mat'].SetValue(name)

    def onSelectMaterial(self, evt=None):
        if self.owids['materials'] is None:
            return
        item = self.owids['materials'].GetSelectedRow()
        name = None
        if item > -1:
            name = list(self.materials_data.keys())[item]
            self.selected_material = name

        self.wids['matrix_btn'].Enable(name is not None)
        if name is not None:
            self.wids['matrix_btn'].SetLabel('Use %s'  % name)

    def onUpdateFilterList(self, evt=None):
        flist = ['None']
        for i in range(len(self.materials_data)):
            if self.owids['materials'].GetToggleValue(i, 3): # is filter
                flist.append(self.owids['materials'].GetTextValue(i, 0))

        for i in range(NFILTERS):
            t = 'filter%d' % (i+1)
            choice = self.wids['%s_mat'%t]
            cur = choice.GetStringSelection()
            choice.Clear()
            choice.SetChoices(flist)
            if cur in flist:
                choice.SetStringSelection(cur)
            else:
                choice.SetSelection(0)

    def onAddMaterial(self, evt=None):
        name    = self.owids['newmat_name'].GetValue()
        formula = self.owids['newmat_form'].GetValue()
        density = self.owids['newmat_dens'].GetValue()
        add = len(name) > 0 and len(formula)>0
        if add and name in self.materials_data:
            add = (Popup(self,
                         "Overwrite definition of '%s'?" % name,
                         'Re-define material?',
                         style=wx.OK|wx.CANCEL)==wx.ID_OK)
            if add:
                irow = list(self.materials_data.keys()).index(name)
                self.owids['materials'].DeleteItem(irow)
        if add:
            add_material(name, formula, density)
            self.materials_data[name] = (formula, density)
            self.selected_material = name
            self.owids['materials'].AppendItem((name, formula,
                                                "%9.6f"%density,
                                                False))

    def onElemSelect(self, event=None, elem=None):
        self.ptable.tsym.SetLabel('')
        self.ptable.title.SetLabel('%d elements selected' %
                                   len(self.ptable.selected))

    def onUsePileupEscape(self, event=None):
        puse = self.wids['pileup_use'].IsChecked()
        self.wids['pileup_amp'].Enable(puse)
        self.wids['pileup_amp_vary'].Enable(puse)

        puse = self.wids['escape_use'].IsChecked()
        self.wids['escape_amp'].Enable(puse)
        self.wids['escape_amp_vary'].Enable(puse)


    def onUsePeak(self, event=None, name=None, value=None):
        if value is None and event is not None:
            value = event.IsChecked()
        if name is None:
            return
        for a in ('cen', 'step', 'tail', 'sigma', 'beta'):
            self.wids['%s_%s'%(name, a)].Enable(value)
            varwid = self.wids.get('%s_%s_vary'%(name, a), None)
            if varwid is not None:
                varwid.Enable(value)

    def build_model(self, match_amplitudes=True):
        """build xrf_model from form settings"""
        vars = {'Vary':'True', 'Fix': 'False', 'True':True, 'False': False}
        opts = {}
        for key, wid in self.wids.items():
            val = None
            if hasattr(wid, 'GetValue'):
                val = wid.GetValue()
            elif hasattr(wid, 'IsChecked'):
                val = wid.IsChecked()
            elif isinstance(wid, Choice):
                val = wid.GetStringSelection()
            elif hasattr(wid, 'GetStringSelection'):
                val = wid.GetStringSelection()
            elif hasattr(wid, 'GetLabel'):
                val = wid.GetLabel()
            if isinstance(val, str) and val.title() in vars:
                val = vars[val.title()]
            opts[key] = val
        opts['count_time'] = getattr(self.mca, 'real_time', 1.0)
        if opts['count_time'] is None:
            opts['count_time'] = 1.0

        script = [xrfmod_setup.format(**opts)]

        for peakname in ('Elastic', 'Compton1', 'Compton2'):
            t = peakname.lower()
            if opts['%s_use'% t]:
                d = {'peakname': t}
                d['_cen']  = opts['%s_cen'%t]
                d['vcen']  = opts['%s_cen_vary'%t]
                d['_step'] = opts['%s_step'%t]
                d['vstep'] = opts['%s_step_vary'%t]
                d['_tail'] = opts['%s_tail'%t]
                d['vtail'] = opts['%s_tail_vary'%t]
                d['_beta'] = opts['%s_beta'%t]
                d['vbeta'] = opts['%s_beta_vary'%t]
                d['_sigma'] = opts['%s_sigma'%t]
                d['vsigma'] = opts['%s_sigma_vary'%t]
                script.append(xrfmod_scattpeak.format(**d))

        for i in range(NFILTERS):
            t = 'filter%d' % (i+1)
            f_mat = opts['%s_mat'%t]
            if f_mat not in (None, 'None') and int(1e6*opts['%s_thk'%t]) > 1:
                script.append(xrfmod_filter.format(name=f_mat,
                                                   thick=opts['%s_thk'%t],
                                                   vary=opts['%s_var'%t]))

        m_mat = opts['matrix_mat'].strip()
        if len(m_mat) > 0 and int(1e6*opts['matrix_thk']) > 1:
            script.append(xrfmod_matrix.format(name=m_mat,
                                               thick=opts['matrix_thk'],
                                               density=opts['matrix_den']))

        if opts['pileup_use'] in ('True', True):
            script.append(xrfmod_pileup.format(scale=opts['pileup_amp'],
                                               vary=opts['pileup_amp_vary']))

        if opts['escape_use'] in ('True', True):
            script.append(xrfmod_escape.format(scale=opts['escape_amp'],
                                               vary=opts['escape_amp_vary']))

        # sort elements selected on Periodic Table by Z
        elemz = []
        for elem in self.ptable.selected:
            elemz.append( 1 + self.ptable.syms.index(elem))
        elemz.sort()
        syms = ["'%s'" % self.ptable.syms[iz-1] for iz in sorted(elemz)]
        syms = '[%s]' % (', '.join(syms))
        script.append(xrfmod_elems.format(elemlist=syms))
        script.append("{group:s}.xrf_init = _xrfmodel.calc_spectrum({group:s}.energy)")
        script = '\n'.join(script)
        self.model_script = script.format(group=self.mcagroup)

        self._larch.eval(self.model_script)

        cmds = []
        self.xrfmod = self._larch.symtable.get_symbol('_xrfmodel')
        floor = 1.e-12*max(self.mca.counts)
        if match_amplitudes:
            total = 0.0 * self.mca.counts
            for name, parr in self.xrfmod.comps.items():
                nam = name.lower()
                try:
                    imax = np.where(parr > 0.99*parr.max())[0][0]
                except:  # probably means all counts are zero
                    imax = int(len(parr)/2.0)
                scale = self.mca.counts[imax] / (parr[imax]+1.00)
                ampname = 'amp_%s' % nam
                if nam in ('elastic', 'compton1', 'compton2', 'compton',
                           'background', 'pileup', 'escape'):
                    ampname = '%s_amp' % nam
                    if nam in ('background', 'pileup', 'escape'):
                        scale = 1.0
                paramval = self.xrfmod.params[ampname].value
                s = "_xrfmodel.params['%s'].value = %.5f" % (ampname, paramval*scale)
                cmds.append(s)
                parr *= scale
                parr[np.where(parr<floor)] = floor
                total += parr
            self.xrfmod.current_model = total
            script = '\n'.join(cmds)
            self._larch.eval(script)
            self.model_script = "%s\n%s" % (self.model_script, script)

        s = "{group:s}.xrf_init = _xrfmodel.calc_spectrum({group:s}.energy)"
        self._larch.eval(s.format(group=self.mcagroup))

    def plot_model(self, model_spectrum=None, init=False, with_comps=True,
                   label=None):
        conf = self.parent.conf

        plotkws = {'linewidth': 2.5, 'delay_draw': True, 'grid': False,
                   'ylog_scale': self.parent.ylog_scale, 'show_legend': False,
                   'fullbox': False}

        ppanel = self.parent.panel
        ppanel.conf.reset_trace_properties()
        self.parent.plot(self.mca.energy, self.mca.counts, mca=self.mca,
                         xlabel='E (keV)', xmin=0, with_rois=False, **plotkws)

        if model_spectrum is None:
            model_spectrum = self.xrfmod.current_model if init else self.xrfmod.best_fit
        if label is None:
            label = 'predicted model' if init else 'best fit'

        self.parent.oplot(self.mca.energy, model_spectrum,
                          label=label, color=conf.fit_color, **plotkws)

        if with_comps:
            for label, arr in self.xrfmod.comps.items():
                ppanel.oplot(self.mca.energy, arr, label=label, **plotkws)

        yscale = {False:'linear', True:'log'}[self.parent.ylog_scale]
        ppanel.set_logscale(yscale=yscale)
        ppanel.set_viewlimits()
        ppanel.conf.set_legend_location('upper right', True)
        ppanel.conf.draw_legend(show=True, delay_draw=False)

    def onShowModel(self, event=None):
        self.build_model()
        self.plot_model(init=True, with_comps=True)

    def onFitIteration(self, iter=0, pars=None):
        pass
        # print("Fit iteration %d" % iter)
        # self.wids['fit_message'].SetLabel("Fit iteration %d" % iter)

    def onFitModel(self, event=None):
        self.build_model()
        xrfmod = self._larch.symtable.get_symbol('_xrfmodel')
        xrfmod.iter_callback = self.onFitIteration
        fit_script = xrfmod_fitscript.format(group=self.mcagroup,
                                             emin=self.wids['en_min'].GetValue(),
                                             emax=self.wids['en_max'].GetValue())

        self._larch.eval(fit_script)
        dgroup = self._larch.symtable.get_group(self.mcagroup)
        self.xrfresults = self._larch.symtable.get_symbol(XRFRESULTS_GROUP)

        xrfresult = self.xrfresults[0]
        xrfresult.script = "%s\n%s" % (self.model_script, fit_script)
        xrfresult.label = "fit %d" % (len(self.xrfresults))
        self.plot_model(init=True, with_comps=True)
        for i in range(len(self.nb.pagelist)):
            if self.nb.GetPageText(i).strip().startswith('Fit R'):
                self.nb.SetSelection(i)
        time.sleep(0.002)
        self.show_results()

    def onClose(self, event=None):
        self.Destroy()

    def onSaveFitResult(self, event=None):
        result = self.get_fitresult()
        deffile = self.mca.label + '_' + result.label
        deffile = fix_filename(deffile.replace('.', '_')) + '.xrfmodel'
        ModelWcards = "XRF Models(*.xrfmodel)|*.xrfmodel|All files (*.*)|*.*"
        sfile = FileSave(self, 'Save XRF Model', default_file=deffile,
                         wildcard=ModelWcards)
        if sfile is not None:
            self._larch.eval(xrfmod_savejs.format(group=self.mcagroup,
                                                  nfit=self.nfit,
                                                  filename=sfile))

    def onExportFitResult(self, event=None):
        result = self.get_fitresult()
        deffile = self.mca.label + '_' + result.label
        deffile = fix_filename(deffile.replace('.', '_')) + '_xrf.txt'
        wcards = 'All files (*.*)|*.*'
        outfile = FileSave(self, 'Export Fit Result', default_file=deffile)
        if outfile is not None:
            buff = ['# XRF Fit %s: %s' % (self.mca.label, result.label),
                    '## Fit Script:']
            for a in result.script.split('\n'):
                buff.append('#   %s' % a)
            buff.append('## Fit Report:')
            for a in result.fit_report.split('\n'):
                buff.append('#   %s' % a)

            buff.append('#')
            buff.append('########################################')

            labels = ['energy', 'counts', 'best_fit',
                      'best_energy', 'fit_window',
                      'fit_weight', 'attenuation']
            labels.extend(list(result.comps.keys()))

            buff.append('# %s' % (' '.join(labels)))

            npts = len(self.mca.energy)
            for i in range(npts):
                dline = [gformat(self.mca.energy[i]),
                         gformat(self.mca.counts[i]),
                         gformat(result.best_fit[i]),
                         gformat(result.best_en[i]),
                         gformat(result.fit_window[i]),
                         gformat(result.fit_weight[i]),
                         gformat(result.atten[i])]
                for c in result.comps.values():
                    dline.append(gformat(c[i]))
                buff.append(' '.join(dline))
            buff.append('\n')
            with open(outfile, 'w') as fh:
                fh.write('\n'.join(buff))

    def get_fitresult(self, nfit=None):
        if nfit is None:
            nfit = self.nfit

        self.xrfresults = self._larch.symtable.get_symbol(XRFRESULTS_GROUP)
        self.nfit = max(0, nfit)
        self.nfit = min(self.nfit, len(self.xrfresults)-1)
        return self.xrfresults[self.nfit]

    def onChangeFitLabel(self, event=None):
        label = self.owids['fitlabel_txt'].GetValue()
        result = self.get_fitresult()
        result.label = label
        self.show_results()

    def onPlot(self, event=None):
        result = self.get_fitresult()
        xrfmod = self._larch.symtable.get_symbol('_xrfmodel')
        with_comps = self.owids['plot_comps'].IsChecked()
        spect = xrfmod.calc_spectrum(self.mca.energy,
                                     params=result.params)
        self.plot_model(model_spectrum=spect, with_comps=with_comps,
                        label=result.label)

    def onSelectFit(self, evt=None):
        if self.owids['stats'] is None:
            return
        item = self.owids['stats'].GetSelectedRow()
        if item > -1:
            self.show_fitresult(nfit=item)

    def onSelectParameter(self, evt=None):
        if self.owids['params'] is None:
            return
        if not self.owids['params'].HasSelection():
            return
        item = self.owids['params'].GetSelectedRow()
        pname = self.owids['paramsdata'][item]

        cormin= self.owids['min_correl'].GetValue()
        self.owids['correl'].DeleteAllItems()

        result = self.get_fitresult()
        this = result.params[pname]
        if this.correl is not None:
            sort_correl = sorted(this.correl.items(), key=lambda it: abs(it[1]))
            for name, corval in reversed(sort_correl):
                if abs(corval) > cormin:
                    self.owids['correl'].AppendItem((pname, name, "% .4f" % corval))

    def onAllCorrel(self, evt=None):
        result = self.get_fitresult()
        params = result.params
        parnames = list(params.keys())

        cormin= self.owids['min_correl'].GetValue()
        correls = {}
        for i, name in enumerate(parnames):
            par = params[name]
            if not par.vary:
                continue
            if hasattr(par, 'correl') and par.correl is not None:
                for name2 in parnames[i+1:]:
                    if (name != name2 and name2 in par.correl and
                            abs(par.correl[name2]) > cormin):
                        correls["%s$$%s" % (name, name2)] = par.correl[name2]

        sort_correl = sorted(correls.items(), key=lambda it: abs(it[1]))
        sort_correl.reverse()

        self.owids['correl'].DeleteAllItems()

        for namepair, corval in sort_correl:
            name1, name2 = namepair.split('$$')
            self.owids['correl'].AppendItem((name1, name2, "% .4f" % corval))

    def show_results(self):
        cur = self.get_fitresult()
        self.owids['stats'].DeleteAllItems()
        for i, res in enumerate(self.xrfresults):
            args = [res.label]
            for attr in ('nvarys', 'nfev', 'chisqr', 'redchi', 'aic'):
                val = getattr(res, attr)
                if isinstance(val, int):
                    val = '%d' % val
                else:
                    val = gformat(val, 11)
                args.append(val)
            self.owids['stats'].AppendItem(tuple(args))
        self.owids['data_title'].SetLabel("%s:  %.3f sec" % (self.mca.label, cur.count_time))
        self.owids['data_title2'].SetLabel("%s:  %.3f sec" % (self.mca.label, cur.count_time))
        self.owids['fitlabel_txt'].SetValue(cur.label)
        self.show_fitresult(nfit=self.nfit)

    def show_fitresult(self, nfit=0, mca=None):
        if mca is not None:
            self.mca = mca
        result = self.get_fitresult(nfit=nfit)

        self.owids['data_title'].SetLabel("%s:  %.3f sec" % (self.mca.label, result.count_time))
        self.owids['data_title2'].SetLabel("%s:  %.3f sec" % (self.mca.label, result.count_time))
        self.result = result
        self.owids['fitlabel_txt'].SetValue(result.label)
        self.owids['params'].DeleteAllItems()
        self.owids['paramsdata'] = []
        for param in reversed(result.params.values()):
            pname = param.name
            try:
                val = gformat(param.value, 10)
            except (TypeError, ValueError):
                val = ' ??? '

            serr, perr = ' N/A ', ' N/A '
            if param.stderr is not None:
                serr = gformat(param.stderr, 10)
                try:
                    perr = ' {:.2%}'.format(abs(param.stderr/param.value))
                except ZeroDivisionError:
                    perr = '?'
            extra = ' '
            if param.expr is not None:
                extra = ' = %s ' % param.expr
            elif not param.vary:
                extra = ' (fixed)'
            elif param.init_value is not None:
                extra = gformat(param.init_value, 10)

            self.owids['params'].AppendItem((pname, val, serr, perr, extra))
            self.owids['paramsdata'].append(pname)
        self.Refresh()



#!/usr/bin/env python
"""
  XAFS Fourier transforms
"""
import numpy as np
from numpy import (pi, arange, zeros, ones, sin, cos,
                   exp, log, sqrt, where, interp, linspace)
# from numpy.fft import fft, ifft
from scipy.fftpack import fft, ifft
from scipy.special import i0 as bessel_i0

from larch import (Group, Make_CallArgs, parse_group_args)

from larch.math import complex_phase
from .xafsutils import set_xafsGroup


MODNAME = '_xafs'
VALID_WINDOWS = ['han', 'fha', 'gau', 'kai', 'par', 'wel', 'sin', 'bes']
sqrtpi = sqrt(pi)

def ftwindow(x, xmin=None, xmax=None, dx=1, dx2=None,
             window='hanning', _larch=None, **kws):
    """
    create a Fourier transform window array.

    Parameters:
    -------------
      x:        1-d array array to build window on.
      xmin:     starting x for FT Window
      xmax:     ending x for FT Window
      dx:       tapering parameter for FT Window
      dx2:      second tapering parameter for FT Window (=dx)
      window:   name of window type

    Returns:
    ----------
    1-d window array.

    Notes:
    -------
    Valid Window names:
        hanning              cosine-squared taper
        parzen               linear taper
        welch                quadratic taper
        gaussian             Gaussian (normal) function window
        sine                 sine function window
        kaiser               Kaiser-Bessel function-derived window

    """
    if window is None:
        window = VALID_WINDOWS[0]
    nam = window.strip().lower()[:3]
    if nam not in VALID_WINDOWS:
        raise RuntimeError("invalid window name %s" % window)

    dx1 = dx
    if dx2 is None:  dx2 = dx1
    if xmin is None: xmin = min(x)
    if xmax is None: xmax = max(x)

    xstep = (x[-1] - x[0]) / (len(x)-1)
    xeps  = 1.e-4 * xstep
    x1 = max(min(x), xmin - dx1/2.0)
    x2 = xmin + dx1/2.0  + xeps
    x3 = xmax - dx2/2.0  - xeps
    x4 = min(max(x), xmax + dx2/2.0)

    if nam == 'fha':
        if dx1 < 0: dx1 = 0
        if dx2 > 1: dx2 = 1
        x2 = x1 + xeps + dx1*(xmax-xmin)/2.0
        x3 = x4 - xeps - dx2*(xmax-xmin)/2.0
    elif nam == 'gau':
        dx1 = max(dx1, xeps)

    def asint(val): return int((val+xeps)/xstep)
    i1, i2, i3, i4 = asint(x1), asint(x2), asint(x3), asint(x4)
    i1, i2 = max(0, i1), max(0, i2)
    i3, i4 = min(len(x)-1, i3), min(len(x)-1, i4)
    if i2 == i1: i1 = max(0, i2-1)
    if i4 == i3: i3 = max(i2, i4-1)
    x1, x2, x3, x4 = x[i1], x[i2], x[i3], x[i4]
    if x1 == x2: x2 = x2+xeps
    if x3 == x4: x4 = x4+xeps
    # initial window
    fwin =  zeros(len(x))
    if i3 > i2:
        fwin[i2:i3] = ones(i3-i2)

    # now finish making window
    if nam in ('han', 'fha'):
        fwin[i1:i2+1] = sin((pi/2)*(x[i1:i2+1]-x1) / (x2-x1))**2
        fwin[i3:i4+1] = cos((pi/2)*(x[i3:i4+1]-x3) / (x4-x3))**2
    elif nam == 'par':
        fwin[i1:i2+1] =     (x[i1:i2+1]-x1) / (x2-x1)
        fwin[i3:i4+1] = 1 - (x[i3:i4+1]-x3) / (x4-x3)
    elif nam == 'wel':
        fwin[i1:i2+1] = 1 - ((x[i1:i2+1]-x2) / (x2-x1))**2
        fwin[i3:i4+1] = 1 - ((x[i3:i4+1]-x3) / (x4-x3))**2
    elif nam  in ('kai', 'bes'):
        cen  = (x4+x1)/2
        wid  = (x4-x1)/2
        arg  = 1 - (x-cen)**2 / (wid**2)
        arg[where(arg<0)] = 0
        if nam == 'bes': # 'bes' : ifeffit 1.0 implementation of kaiser-bessel
            fwin = bessel_i0(dx* sqrt(arg)) / bessel_i0(dx)
            fwin[where(x<=x1)] = 0
            fwin[where(x>=x4)] = 0
        else: # better version
            scale = max(1.e-10, bessel_i0(dx)-1)
            fwin = (bessel_i0(dx * sqrt(arg)) - 1) / scale
    elif nam == 'sin':
        fwin[i1:i4+1] = sin(pi*(x4-x[i1:i4+1]) / (x4-x1))
    elif nam == 'gau':
        cen  = (x4+x1)/2
        fwin =  exp(-(((x - cen)**2)/(2*dx1*dx1)))
    return fwin


@Make_CallArgs(["r", "chir"])
def xftr(r, chir=None, group=None, rmin=0, rmax=20, with_phase=False,
            dr=1, dr2=None, rw=0, window='kaiser', qmax_out=None,
            nfft=2048, kstep=0.05, _larch=None, **kws):
    """
    reverse XAFS Fourier transform, from chi(R) to chi(q).

    calculate reverse XAFS Fourier transform
    This assumes that chir_re and (optional chir_im are
    on a uniform r-grid given by r.

    Parameters:
    ------------
      r:        1-d array of distance, or group.
      chir:     1-d array of chi(R)
      group:    output Group
      qmax_out: highest *k* for output data (30 Ang^-1)
      rweight:  exponent for weighting spectra by r^rweight (0)
      rmin:     starting *R* for FT Window
      rmax:     ending *R* for FT Window
      dr:       tapering parameter for FT Window
      dr2:      second tapering parameter for FT Window
      window:   name of window type
      nfft:     value to use for N_fft (2048).
      kstep:    value to use for delta_k (0.05).
      with_phase: output the phase as well as magnitude, real, imag  [False]

    Returns:
    ---------
      None -- outputs are written to supplied group.

    Notes:
    -------
    Arrays written to output group:
        rwin               window Omega(R) (length of input chi(R)).
        q                  uniform array of k, out to qmax_out.
        chiq               complex array of chi(k).
        chiq_mag           magnitude of chi(k).
        chiq_re            real part of chi(k).
        chiq_im            imaginary part of chi(k).
        chiq_pha           phase of chi(k) if with_phase=True
                           (a noticable performance hit)

    Supports First Argument Group convention (with group member names 'r' and 'chir')
    """
    if 'rweight' in kws:
        rw = kws['rweight']

    r, chir, group = parse_group_args(r, members=('r', 'chir'),
                                     defaults=(chir,), group=group,
                                     fcn_name='xftr')
    rstep = r[1] - r[0]
    kstep = pi/(rstep*nfft)
    scale = 1.0

    cchir = zeros(nfft, dtype='complex128')
    r_    = rstep * arange(nfft, dtype='float64')

    cchir[0:len(chir)] = chir
    if chir.dtype == np.dtype('complex128'):
        scale = 0.5

    win = ftwindow(r_, xmin=rmin, xmax=rmax, dx=dr, dx2=dr2, window=window)
    out = scale * xftr_fast( cchir*win * r_**rw, kstep=kstep, nfft=nfft)
    if qmax_out is None: qmax_out = 30.0
    q = linspace(0, qmax_out, int(1.05 + qmax_out/kstep))
    nkpts = len(q)

    group = set_xafsGroup(group, _larch=_larch)
    group.q = q
    mag = sqrt(out.real**2 + out.imag**2)
    group.rwin =  win[:len(chir)]
    group.chiq     =  out[:nkpts]
    group.chiq_mag =  mag[:nkpts]
    group.chiq_re  =  out.real[:nkpts]
    group.chiq_im  =  out.imag[:nkpts]
    if with_phase:
        group.chiq_pha =  complex_phase(out[:nkpts])



@Make_CallArgs(["k", "chi"])
def xftf(k, chi=None, group=None, kmin=0, kmax=20, kweight=0,
         dk=1, dk2=None, with_phase=False, window='kaiser', rmax_out=10,
         nfft=2048, kstep=0.05, _larch=None, **kws):
    """
    forward XAFS Fourier transform, from chi(k) to chi(R), using
    common XAFS conventions.

    Parameters:
    -----------
      k:        1-d array of photo-electron wavenumber in Ang^-1 or group
      chi:      1-d array of chi
      group:    output Group
      rmax_out: highest R for output data (10 Ang)
      kweight:  exponent for weighting spectra by k**kweight
      kmin:     starting k for FT Window
      kmax:     ending k for FT Window
      dk:       tapering parameter for FT Window
      dk2:      second tapering parameter for FT Window
      window:   name of window type
      nfft:     value to use for N_fft (2048).
      kstep:    value to use for delta_k (0.05 Ang^-1).
      with_phase: output the phase as well as magnitude, real, imag  [False]

    Returns:
    ---------
      None   -- outputs are written to supplied group.

    Notes:
    -------
    Arrays written to output group:
        kwin               window function Omega(k) (length of input chi(k)).
        r                  uniform array of R, out to rmax_out.
        chir               complex array of chi(R).
        chir_mag           magnitude of chi(R).
        chir_re            real part of chi(R).
        chir_im            imaginary part of chi(R).
        chir_pha           phase of chi(R) if with_phase=True
                           (a noticable performance hit)

    Supports First Argument Group convention (with group member names 'k' and 'chi')
    """
    # allow kweight keyword == kw
    if 'kw' in kws:
        kweight = kws['kw']

    k, chi, group = parse_group_args(k, members=('k', 'chi'),
                                     defaults=(chi,), group=group,
                                     fcn_name='xftf')

    cchi, win  = xftf_prep(k, chi, kmin=kmin, kmax=kmax, kweight=kweight,
                               dk=dk, dk2=dk2, nfft=nfft, kstep=kstep,
                               window=window, _larch=_larch)

    out = xftf_fast(cchi*win, kstep=kstep, nfft=nfft)
    rstep = pi/(kstep*nfft)

    irmax = int(min(nfft/2, 1.01 + rmax_out/rstep))

    group = set_xafsGroup(group, _larch=_larch)
    r   = rstep * arange(irmax)
    mag = sqrt(out.real**2 + out.imag**2)
    group.kwin =  win[:len(chi)]
    group.r    =  r[:irmax]
    group.chir =  out[:irmax]
    group.chir_mag =  mag[:irmax]
    group.chir_re  =  out.real[:irmax]
    group.chir_im  =  out.imag[:irmax]
    if with_phase:
        group.chir_pha =  complex_phase(out[:irmax])



def xftf_prep(k, chi, kmin=0, kmax=20, kweight=2, dk=1, dk2=None,
                window='kaiser', nfft=2048, kstep=0.05, _larch=None):
    """
    calculate weighted chi(k) on uniform grid of len=nfft, and the
    ft window.

    Returns weighted chi, window function which can easily be multiplied
    and used in xftf_fast.
    """
    if dk2 is None: dk2 = dk
    npts = int(1.01 + max(k)/kstep)
    k_max = max(max(k), kmax+dk2)
    k_   = kstep * np.arange(int(1.01+k_max/kstep), dtype='float64')
    chi_ = interp(k_, k, chi)
    win  = ftwindow(k_, xmin=kmin, xmax=kmax, dx=dk, dx2=dk2, window=window)
    return ((chi_[:npts] *k_[:npts]**kweight), win[:npts])


def xftf_fast(chi, nfft=2048, kstep=0.05, _larch=None, **kws):
    """
    calculate forward XAFS Fourier transform.  Unlike xftf(),
    this assumes that:
      1. data is already on a uniform grid
      2. any windowing and/or kweighting has been applied.
    and simply returns the complex chi(R), not setting any larch data.

    This is useful for repeated FTs, as inside loops.

    Parameters:
    ------------
      chi:      1-d array of chi to be transformed
      nfft:     value to use for N_fft (2048).
      kstep:    value to use for delta_k (0.05).

    Returns:
    --------
      complex 1-d array chi(R)

    """
    cchi = zeros(nfft, dtype='complex128')
    cchi[0:len(chi)] = chi
    return (kstep / sqrtpi) * fft(cchi)[:int(nfft/2)]

def xftr_fast(chir, nfft=2048, kstep=0.05, _larch=None, **kws):
    """
    calculate reverse XAFS Fourier transform, from chi(R) to
    chi(q), using common XAFS conventions.  This version demands
    chir be the complex chi(R) as created from xftf().

    It returns the complex array of chi(q) without putting any
    values into an output group.

    Parameters:
    -------------
      chir:     1-d array of chi(R) to be transformed
      nfft:     value to use for N_fft (2048).
      kstep:    value to use for delta_k (0.05).

    Returns:
    ----------
      complex 1-d array for chi(q).

    This is useful for repeated FTs, as inside loops.
    """
    cchi = zeros(nfft, dtype='complex128')
    cchi[0:len(chir)] = chir
    return  (4*sqrtpi/kstep) * ifft(cchi)[:int(nfft/2)]



#!/usr/bin/env python
"""
GUI for displaying maps from HDF5 files

"""

VERSION = '10 (14-March-2018)'

import os
import platform
import sys
import time
import json
import socket
import datetime
from functools import partial
from threading import Thread

import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.mixins.inspection
try:
    from wx._core import PyDeadObjectError
except:
    PyDeadObjectError = Exception

HAS_tomopy = False
try:
    import tomopy
    HAS_tomopy = True
except ImportError:
    pass

import numpy as np
import scipy.stats as stats

from  ..wxlib import (EditableListBox, SimpleText,
                       FloatCtrl, Font, pack, Popup, Button, MenuItem,
                       Choice, Check, GridPanel, FileSave, HLine)
from ..utils.strutils import bytes2str, version_ge
from ..io import nativepath
from ..math.tomography import TOMOPY_ALG, TOMOPY_FILT

from ..xrmmap import GSEXRM_MapFile, GSEXRM_FileStatus, h5str, ensure_subgroup


CEN = wx.ALIGN_CENTER
LEFT = wx.ALIGN_LEFT
RIGHT = wx.ALIGN_RIGHT
ALL_CEN =  wx.ALL|CEN
ALL_LEFT =  wx.ALL|LEFT
ALL_RIGHT =  wx.ALL|RIGHT


PLOT_TYPES = ('Single ROI Map', 'Three ROI Map', 'Correlation Plot')
PLOT_OPERS = ('/', '*', '-', '+')
CONTRAST_CHOICES = ('None',
                    '0.01', '0.02', '0.05',
                    '0.1', '0.2', '0.5',
                    '1', '2', '5')

CWID = 150
WWID = 100 + CWID*4

class TomographyPanel(GridPanel):
    '''Panel of Controls for reconstructing a tomographic slice'''
    label  = 'Tomography Tools'
    def __init__(self, parent, owner=None, **kws):

        self.owner = owner
        self.cfile,self.xrmmap = None,None
        self.npts = None
        self.resave = False

        GridPanel.__init__(self, parent, nrows=8, ncols=6, **kws)

        self.plot_choice = Choice(self, choices=PLOT_TYPES[:-1], size=(CWID, -1))
        self.plot_choice.Bind(wx.EVT_CHOICE, self.plotSELECT)

        self.det_choice = [Choice(self, size=(CWID, -1)),
                           Choice(self, size=(CWID, -1)),
                           Choice(self, size=(CWID, -1)),
                           Choice(self, size=(CWID, -1))]
        self.roi_choice = [Choice(self, size=(CWID, -1)),
                           Choice(self, size=(CWID, -1)),
                           Choice(self, size=(CWID, -1)),
                           Choice(self, size=(CWID, -1))]

        fopts = dict(minval=0, precision=2, size=(110, -1))
        self.iminvals = [FloatCtrl(self, value=0, **fopts),
                         FloatCtrl(self, value=0, **fopts),
                         FloatCtrl(self, value=0, **fopts)]
        self.imaxvals = [FloatCtrl(self, value=0, **fopts),
                         FloatCtrl(self, value=0, **fopts),
                         FloatCtrl(self, value=0, **fopts)]
        self.icontrast = [Choice(self, choices=CONTRAST_CHOICES, default=4, size=(CWID, -1)),
                          Choice(self, choices=CONTRAST_CHOICES, default=4, size=(CWID, -1)),
                          Choice(self, choices=CONTRAST_CHOICES, default=4, size=(CWID, -1))]

        for i, d in enumerate(self.icontrast):
            d.Bind(wx.EVT_CHOICE, partial(self.roiContrast, i))

        for i,det_chc in enumerate(self.det_choice):
            det_chc.Bind(wx.EVT_CHOICE, partial(self.detSELECT,i))

        for i,roi_chc in enumerate(self.roi_choice):
            roi_chc.Bind(wx.EVT_CHOICE, partial(self.roiSELECT,i))

        self.det_label = [SimpleText(self,'Intensity'),
                          SimpleText(self,''),
                          SimpleText(self,''),
                          SimpleText(self,'Normalization')]
        self.roi_label = [SimpleText(self,''),
                          SimpleText(self,''),
                          SimpleText(self,''),
                          SimpleText(self,'')]

        self.use_dtcorr  = Check(self, default=True,
                                 label='Correct for Detector Deadtime',
                                 action=self.onDTCorrect)
        self.use_hotcols = Check(self, default=False,
                                 label='Remove First and Last columns',
                                 action=self.onHotCols)
        self.i1trans = Check(self, default=True,
                             label='Scalar "i1" is transmission data')

        self.tomo_show = [Button(self, 'Show New Map',     size=(CWID, -1),
                               action=partial(self.onShowTomograph, new=True)),
                          Button(self, 'Replace Last Map', size=(CWID, -1),
                               action=partial(self.onShowTomograph, new=False))]

        self.tomo_algo = Choice(self, choices=TOMOPY_ALG, size=(CWID, -1),
                                action=self.onALGchoice)
        self.tomo_filt = Choice(self, choices=TOMOPY_FILT, size=(CWID, -1))
        self.tomo_niter = wx.SpinCtrl(self, min=1, max=500, initial=1,
                                      size=(CWID, -1),
                                      style=wx.SP_VERTICAL|wx.SP_ARROW_KEYS|wx.SP_WRAP)

        self.center_value = wx.SpinCtrlDouble(self, inc=0.25, size=(100, -1),
                                     style=wx.SP_VERTICAL|wx.SP_ARROW_KEYS|wx.SP_WRAP)
        self.center_value.SetIncrement(0.25)
        self.center_value.SetDigits(2)
        self.refine_center = wx.CheckBox(self, label='Refine')
        self.refine_center.SetValue(False)

        self.sino_data   = Choice(self, size=(200, -1))
        self.tomo_save   = Button(self, 'Save reconstruction',     size=(150, -1),
                               action=self.onSaveTomograph)


        #################################################################################

        self.Add(SimpleText(self, 'Display Virtual Slices:    Plot Type:'), dcol=2,
                 style=LEFT, newrow=True)

        self.Add(self.plot_choice, dcol=1, style=LEFT)
        self.Add(self.i1trans,     dcol=2, style=LEFT)
        self.Add(SimpleText(self,'Options:'),   dcol=1, style=LEFT, newrow=True)
        self.Add(self.use_dtcorr,               dcol=2, style=LEFT)
        self.Add(self.use_hotcols,              dcol=2, style=LEFT)

        self.AddMany((SimpleText(self,''), self.det_label[0],
                        self.det_label[1], self.det_label[2], self.det_label[3]),
                     style=LEFT,  newrow=True)

        self.AddMany((SimpleText(self,'Detector:'), self.det_choice[0],
                      self.det_choice[1], self.det_choice[2], self.det_choice[3]),
                      style=LEFT,  newrow=True)

        self.AddMany((SimpleText(self,'ROI:'), self.roi_choice[0],
                      self.roi_choice[1], self.roi_choice[2], self.roi_choice[3]),
                     style=LEFT,  newrow=True)

        self.AddMany((SimpleText(self,''), self.roi_label[0],
                      self.roi_label[1], self.roi_label[2],
                      self.roi_label[3]), style=LEFT, newrow=True)

        self.AddMany((SimpleText(self,'I Min:'), self.iminvals[0],
                      self.iminvals[1], self.iminvals[2]), style=LEFT,
                     newrow=True)

        self.AddMany((SimpleText(self,'I Max:'), self.imaxvals[0],
                      self.imaxvals[1], self.imaxvals[2]), style=LEFT,
                     newrow=True)

        self.AddMany((SimpleText(self,'I contrast %:'), self.icontrast[0],
                      self.icontrast[1], self.icontrast[2]), style=LEFT,
                     newrow=True)


        self.Add((5, 5),                        dcol=1, style=LEFT, newrow=True)
        self.Add((5, 5),                        dcol=1, style=LEFT, newrow=True)
        self.Add(self.tomo_show[0],             dcol=1, style=LEFT)
        self.Add(self.tomo_show[1],             dcol=1, style=LEFT)

        self.Add(HLine(self, size=(WWID, 5)),    dcol=8, style=LEFT,  newrow=True)

        self.Add(SimpleText(self,'Reconstruction '), dcol=2, style=LEFT,  newrow=True)

        self.Add(SimpleText(self,'Algorithm:'),     dcol=1, style=LEFT, newrow=True)
        self.Add(self.tomo_algo,                    dcol=1, style=LEFT)
        self.Add(SimpleText(self,'Filter: '),       dcol=1, style=LEFT)
        self.Add(self.tomo_filt,                    dcol=1, style=LEFT)

        self.Add(SimpleText(self,'# Iterations'),   dcol=1, style=LEFT, newrow=True)
        self.Add(self.tomo_niter,                   dcol=1, style=LEFT)
        self.Add(SimpleText(self,'Center Pixel:'),      dcol=1, style=LEFT)
        self.Add(self.center_value, dcol=1, style=LEFT)
        self.Add(self.refine_center, dcol=1, style=LEFT)

        self.Add(HLine(self, size=(WWID, 5)),     dcol=8, style=LEFT,  newrow=True)


        self.Add(SimpleText(self,'Data:'),             dcol=1, style=LEFT,  newrow=True)
        self.Add(self.sino_data,                       dcol=2, style=LEFT)
        self.Add(self.tomo_save,                       dcol=2, style=LEFT)

        #################################################################################
        self.pack()

    def onDTCorrect(self, event=None):
        self.owner.current_file.dtcorrect = self.use_dtcorr.IsChecked()

    def onHotCols(self, event=None):
        self.owner.current_file.hotcols = self.use_hotcols.IsChecked()

    def update_xrmmap(self, xrmfile=None, set_detectors=None):

        if xrmfile is None:
            xrmfile = self.owner.current_file

        self.cfile  = xrmfile
        self.xrmmap = self.cfile.xrmmap


        if self.cfile.get_rotation_axis() is None:
            self.center_value.SetValue(0)
            return

        self.set_det_choices()

        try:
            self.npts = len(self.cfile.get_pos(0, mean=True))
        except:
            self.npts = len(self.cfile.get_pos('x', mean=True))

        center = self.cfile.get_tomography_center()
        self.center_value.SetRange(-0.5*self.npts,1.5*self.npts)
        self.center_value.SetValue(center)

        self.plotSELECT()


    def onALGchoice(self,event=None):

        alg = self.tomo_algo.GetStringSelection().lower()
        enable_filter = False
        enable_niter = False

        if alg.startswith('gridrec'):
            enable_filter = True
        else:
            enable_niter = True

        self.tomo_niter.Enable(enable_niter)
        self.tomo_filt.Enable(enable_filter)

    def detSELECT(self, idet, event=None):
        self.set_roi_choices(idet=idet)

    def roiContrast(self, iroi, event=None):
        if iroi > 2:
            return
        try:
            detname = self.det_choice[iroi].GetStringSelection()
            roiname = self.roi_choice[iroi].GetStringSelection()
            contrast = self.icontrast[iroi].GetStringSelection()
        except:
            return
        if contrast in ('None', None):
            contrast = 0.0
        contrast = float(contrast)
        try:
            map = self.cfile.get_roimap(roiname, det=detname)
            imin, imax = np.percentile(map, (contrast, 100.0-contrast))
            self.iminvals[iroi].SetValue(imin)
            self.imaxvals[iroi].SetValue(imax)
        except:
            pass

    def roiSELECT(self, iroi, event=None):
        detname = self.det_choice[iroi].GetStringSelection()
        roiname = self.roi_choice[iroi].GetStringSelection()
        try:
            contrast = self.icontrast[iroi].GetStringSelection()
        except:
            contrast = 0.0
        if contrast in ('None', None):
            contrast = 0.0
        contrast = float(contrast)


        if version_ge(self.cfile.version, '2.0.0'):
            try:
                roi = self.cfile.xrmmap['roimap'][detname][roiname]
                limits = roi['limits'][:]
                units = bytes2str(roi['limits'].attrs.get('units',''))
                if units == '1/A':
                    roistr = '[%0.2f to %0.2f %s]' % (limits[0],limits[1],units)
                else:
                    roistr = '[%0.1f to %0.1f %s]' % (limits[0],limits[1],units)
            except:
                roistr = ''
            try:
                map = self.cfile.get_roimap(roiname, det=detname)
                imin, imax = np.percentile(map, (contrast, 100.0-contrast))
                self.iminvals[iroi].SetValue(imin)
                self.imaxvals[iroi].SetValue(imax)
            except:
                pass
        else:
            try:
                roi = self.cfile.xrmmap[detname]
                en     = list(roi['energy'][:])
                index  = list(roi['roi_name'][:]).index(roiname)
                limits = list(roi['roi_limits'][:][index])
                roistr = '[%0.1f to %0.1f keV]' % (en[limits[0]],en[limits[1]])
            except:
                roistr = ''

        self.roi_label[iroi].SetLabel(roistr)

    def plotSELECT(self,event=None):
        if len(self.owner.filemap) > 0:
            plot_type = self.plot_choice.GetStringSelection().lower()
            if 'single' in plot_type:
                for i in (1,2):
                    self.det_choice[i].Disable()
                    self.roi_choice[i].Disable()
                    self.roi_label[i].SetLabel('')
                for i,label in enumerate(['Intensity', ' ', ' ']):
                    self.det_label[i].SetLabel(label)
            elif 'three' in plot_type:
                for i in (1,2):
                    self.det_choice[i].Enable()
                    self.roi_choice[i].Enable()
                for i,label in enumerate(['Red', 'Green', 'Blue']):
                    self.det_label[i].SetLabel(label)
                self.set_roi_choices()

    def onLasso(self, selected=None, mask=None, data=None, xrmfile=None, **kws):
        if xrmfile is None: xrmfile = self.owner.current_file
        ny, nx = xrmfile.get_shape()
        indices = []
        for idx in selected:
            iy, ix = divmod(idx, ny)
            indices.append((ix, iy))


    def onClose(self):
        for p in self.plotframes:
            try:
                p.Destroy()
            except:
                pass

    def calculateSinogram(self,xrmfile=None):
        '''
        returns slice as [slices, x, 2th]
        '''
        subtitles = None
        plt3 = 'three' in self.plot_choice.GetStringSelection().lower()

        det_name = ['mcasum']*4
        roi_name = ['']*4
        plt_name = ['']*4
        minvals  = [0]*4
        maxvals = [np.inf]*4
        for i in range(4):
            det_name[i] = self.det_choice[i].GetStringSelection()
            roi_name[i] = self.roi_choice[i].GetStringSelection()
            if det_name[i] == 'scalars':
                plt_name[i] = '%s' % roi_name[i]
            else:
                plt_name[i] = '%s(%s)' % (roi_name[i],det_name[i])
            if i < 3:
                minvals[i] = self.iminvals[i].GetValue()
                maxvals[i] = self.imaxvals[i].GetValue()

        if plt3:
            flagxrd = False
            for det in det_name:
                if det.startswith('xrd'): flagxrd = True
        else:
            flagxrd = True if det_name[0].startswith('xrd') else False

        if xrmfile is None:
            xrmfile = self.owner.current_file

        args={'trim_sino' : flagxrd,
              'hotcols'   : False,
              'dtcorrect' : self.owner.dtcor}

        x     = xrmfile.get_translation_axis(hotcols=args['hotcols'])
        omega = xrmfile.get_rotation_axis(hotcols=args['hotcols'])

        if omega is None:
            print('\n** Cannot compute tomography: no rotation axis specified in map. **')
            return

        # check for common case of a few too many angles -- in which case, always
        # remove the first and last:
        domega  = abs(np.diff(omega).mean())
        if abs(omega[-1] - omega[0]) > 360+2*domega:
            omega = omega[1:-1]
            args['hotcols'] = True

        def normalize_map(xmap, normmap, roiname):
            xmap /= normmap
            label = ''
            if self.i1trans.IsChecked() and roiname.lower().startswith('i1'):
                xmap = -np.log(xmap)
                xmrange = xmap.max()-xmap.min()
                xmap = (xmap - xmap.min() + 1.e-6*xmrange)/xmrange
                label = '-log'
            return xmap, label

        normmap = 1.
        if roi_name[-1] != '1':
            normmap, sino_order = xrmfile.get_sinogram(roi_name[-1],
                                                       det=det_name[-1], **args)
            normmap[np.where(normmap==0)] = 1.

        r_map, sino_order = xrmfile.get_sinogram(roi_name[0],
                                                 det=det_name[0],
                                                 minval=minvals[0],
                                                 maxval=maxvals[0], **args)
        r_map, r_lab = normalize_map(r_map, normmap, roi_name[0])
        if plt3:
            g_map, sino_order = xrmfile.get_sinogram(roi_name[1], det=det_name[1],
                                                     minval=minvals[1],
                                                     maxval=maxvals[1], **args)
            b_map, sino_order = xrmfile.get_sinogram(roi_name[2], det=det_name[2],
                                                     minval=minvals[2],
                                                     maxval=maxvals[2], **args)
            g_map, g_lab = normalize_map(g_map, normmap, roi_name[1])
            b_map, b_lab = normalize_map(b_map, normmap, roi_name[2])


        pref, fname = os.path.split(xrmfile.filename)
        if plt3:
            sino = np.array([r_map, g_map, b_map])
            sino.resize(tuple(i for i in sino.shape if i!=1))
            title = fname
            info = ''
            if roi_name[-1] == '1':
                subtitles = {'red':   'Red: %s'   % plt_name[0],
                             'green': 'Green: %s' % plt_name[1],
                             'blue':  'Blue: %s'  % plt_name[2]}
            else:
                subtitles = {'red':   'Red: %s(%s/%s)'   % (r_lab, plt_name[0], plt_name[-1]),
                             'green': 'Green: %s(%s/%s)' % (g_lab, plt_name[1], plt_name[-1]),
                             'blue':  'Blue: %s(%s/%s)'  % (b_lab, plt_name[2], plt_name[-1])}

        else:
            sino = r_map
            if roi_name[-1] == '1':
                title = plt_name[0]
            else:
                title = '%s(%s/%s)' % (r_lab, plt_name[0] , plt_name[-1])
            title = '%s: %s' % (fname, title)
            info  = 'Intensity: [%g, %g]' %(sino.min(), sino.max())
            subtitle = None

        return title, subtitles, info, x, omega, sino_order, sino

    def onSaveTomograph(self, event=None):

        xrmfile = self.owner.current_file
        detpath = self.sino_data.GetStringSelection()
        center = self.center_value.GetValue()

        if not self.owner.dtcor and 'scalars' in detpath:
            detpath = '%s_raw' % detpath

        print('\nSaving tomographic reconstruction for %s ...' % detpath)

        xrmfile.save_tomograph(detpath,
                               algorithm=self.tomo_algo.GetStringSelection(),
                               filter_name=self.tomo_filt.GetStringSelection(),
                               num_iter=self.tomo_niter.GetValue(),
                               center=center, dtcorrect=self.owner.dtcor,
                               hotcols=xrmfile.hotcols)
        print('Saved.')


    def onShowTomograph(self, event=None, new=True):
        xrmfile = self.owner.current_file
        det = None
        title, subtitles, info, x, omega, sino_order, sino = self.calculateSinogram()

        algorithm = self.tomo_algo.GetStringSelection()
        filter_name = self.tomo_filt.GetStringSelection()
        niter = self.tomo_niter.GetValue()
        center = self.center_value.GetValue()
        refine_center = self.refine_center.GetValue()

        tomo = xrmfile.get_tomograph(sino, refine_center=refine_center,
                                     algorithm=algorithm,
                                     filter_name=filter_name, num_iter=niter,
                                     center=center, omega=omega,
                                     sinogram_order=sino_order,
                                     hotcols=xrmfile.hotcols)

        # sharpness estimates:
        if len(tomo.shape) == 3:
            t = tomo.sum(axis=2)/tomo.max()
        else:
            t = tomo/tomo.max()

        _mean = ((t-t.mean())**2).mean()
        hist, _ = np.histogram(t, bins=128, range=[t.min(), t.max()])
        hist = hist.astype('float64')/t.size
        hist[np.where(hist<1.e-15)] = 1.e-15
        _negent = -np.dot(hist, np.log(hist))
        # print("sharpness center=%f  mean=%g ent=%g" % (center, _mean, _negent))


        if refine_center:
            self.set_center(xrmfile.xrmmap['tomo/center'].value)
            self.refine_center.SetValue(False)



        omeoff, xoff = 0, 0
        title = '%s, center=%0.1f' % (title, center)

        ## for one color plot
        if sino.shape[0] == 1 and tomo.shape[0] == 1:
            sino = sino[0]
            tomo = tomo[0]
            det = self.det_choice[0].GetStringSelection()

        if len(self.owner.tomo_displays) == 0 or new:
            iframe = self.owner.add_tomodisplay(title)
        self.owner.display_tomo(tomo, title=title, subtitles=subtitles, det=det)

    def set_center(self,cen):
        self.center_value.SetValue(cen)
        self.cfile.set_tomography_center(center=cen)

    def set_det_choices(self):
        det_list = self.cfile.get_detector_list()

        for det_ch in self.det_choice:
            det_ch.SetChoices(det_list)
        if 'scalars' in det_list: ## should set 'denominator' to scalars as default
            self.det_choice[-1].SetStringSelection('scalars')

        data_list = self.cfile.get_datapath_list(remove='raw')
        self.sino_data.SetChoices(data_list)

        self.set_roi_choices()

    def set_roi_choices(self, idet=None):

        if idet is None:
            for idet,det_ch in enumerate(self.det_choice):
                detname = self.det_choice[idet].GetStringSelection()
                rois = self.update_roi(detname)

                self.roi_choice[idet].SetChoices(rois)
                self.roiSELECT(idet)
        else:
            detname = self.det_choice[idet].GetStringSelection()
            rois = self.update_roi(detname)

            self.roi_choice[idet].SetChoices(rois)
            self.roiSELECT(idet)


    def update_roi(self, detname):
        return self.cfile.get_roi_list(detname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
RIXS profile toolbar
====================

A modified version of the SILX profile toolbar.
"""
from itertools import cycle
from silx.gui.plot.Profile import (ProfileToolBar, createProfile)


_DEFAULT_OVERLAY_COLORS = cycle(['#1F77B4', '#AEC7E8', '#FF7F0E', '#FFBB78',
                                 '#2CA02C', '#98DF8A', '#D62728', '#FF9896',
                                 '#9467BD', '#C5B0D5', '#8C564B', '#C49C94',
                                 '#E377C2', '#F7B6D2', '#7F7F7F', '#C7C7C7',
                                 '#BCBD22', '#DBDB8D', '#17BECF', '#9EDAE5'])


class RixsProfileToolBar(ProfileToolBar):
    """RIXS-adapted Profile (=Cuts) toolbar"""

    def __init__(self, parent=None, plot=None, profileWindow=None,
                 overlayColors=None, title='RIXS profile'):
        """Constructor"""
        super(RixsProfileToolBar, self).__init__(parent=parent, plot=plot,
                                                 profileWindow=profileWindow,
                                                 title=title)

        self._overlayColors = overlayColors or _DEFAULT_OVERLAY_COLORS

    def _getNewColor(self):
        return next(self._overlayColors)

    def updateProfile(self):
        """Update the displayed profile and profile ROI.
        This uses the current active image of the plot and the current ROI.
        """
        image = self.plot.getActiveImage()
        if image is None:
            return

        self._overlayColor = self._getNewColor()

        self._createProfile(currentData=image.getData(copy=False),
                            origin=image.getOrigin(), scale=image.getScale(),
                            colormap=None, z=image.getZValue(),
                            method=self.getProfileMethod())


    def _createProfile(self, currentData, origin, scale, colormap, z, method):
        """Create the profile line for the the given image.
        :param numpy.ndarray currentData: the image or the stack of images
            on which we compute the profile
        :param origin: (ox, oy) the offset from origin
        :type origin: 2-tuple of float
        :param scale: (sx, sy) the scale to use
        :type scale: 2-tuple of float
        :param dict colormap: The colormap to use
        :param int z: The z layer of the image
        """
        if self._roiInfo is None:
            return

        coords, profile, area, profileName, xLabel = createProfile(
            roiInfo=self._roiInfo,
            currentData=currentData,
            origin=origin,
            scale=scale,
            lineWidth=self.lineWidthSpinBox.value(),
            method=method)

        profilePlot = self.getProfilePlot()
        plotTitle = self.plot.getGraphTitle()

        profilePlot.setGraphTitle("Profiles")
        profilePlot.getXAxis().setLabel(xLabel)

        profileName = "{0}: {1}".format(plotTitle, profileName)

        dataIs3D = len(currentData.shape) > 2
        if dataIs3D:
            profileScale = (coords[-1] - coords[0]) / profile.shape[1], 1
            profilePlot.addImage(profile,
                                 legend=profileName,
                                 colormap=colormap,
                                 origin=(coords[0], 0),
                                 scale=profileScale)
            profilePlot.getYAxis().setLabel("Frame index (depth)")
        else:
            profilePlot.addCurve(coords,
                                 profile[0],
                                 legend=profileName,
                                 color=self.overlayColor)

        self.plot.addItem(area[0], area[1],
                          legend=self._POLYGON_LEGEND,
                          color=self.overlayColor,
                          shape='polygon', fill=True,
                          replace=False, z=z + 1)

        self._showProfileMainWindow()




if __name__ == '__main__':
    pass




import json
import time
import numpy as np

from collections import OrderedDict

from larch import Group
from ..fitting import Parameter, isParameter
from ..utils.jsonutils import encode4js, decode4js
from . import fix_varname


def save(fname,  *args, _larch=None, **kws):
    """save groups and data into a portable json file

    save(fname, arg1, arg2, ....)

    Parameters
    ----------
       fname   name of output save file.
       args    list of groups, data items to be saved.

    See Also:  restore()
    """
    isgroup =  _larch.symtable.isgroup

    expr = getattr(_larch, 'this_expr', 'save(foo)')
    expr = expr.replace('\n', ' ').replace('\r', ' ')

    grouplist = _larch.symtable._sys.saverestore_groups[:]

    buff = ["#Larch Save File: 1.0",
            "#save.date: %s" % time.strftime('%Y-%m-%d %H:%M:%S'),
            "#save.command: %s" % expr,
            "#save.nitems:  %i" % len(args)]

    names = []
    if expr.startswith('save('):
        names = [a.strip() for a in expr[5:-1].split(',')]
    try:
        names.pop(0)
    except:
        pass
    if len(names) < len(args):
        names.extend(["_unknown_"]*(len(args) - len(names)))

    for name, arg in zip(names, args):
        buff.append("#=> %s" % name)
        buff.append(json.dumps(encode4js(arg, grouplist=grouplist)))
    buff.append("")
    with open(fname, "w") as fh:
        fh.write("\n".join(buff))



def restore(fname, top_level=True, _larch=None):
    """restore data from a json Larch save file

    Arguments
    ---------
    top_level  bool  whether to restore to _main [True]


    Returns
    -------
    None   with `top_level=True` or group with `top_level=False`

    Notes
    -----
    1.  With top_level=False, a new group containing the
        recovered data will be returned.
    """

    grouplist = _larch.symtable._sys.saverestore_groups

    datalines = open(fname, 'r').readlines()
    line1 = datalines.pop(0)
    if not line1.startswith("#Larch Save File:"):
        raise ValueError("%s is not a valid Larch save file" % fname)
    version_string = line1.split(':')[1].strip()
    version_info = [s for s in version_string.split('.')]

    ivar = 0
    header = {'version': version_info}
    varnames = []
    gname = fix_varname('restore_%s' % fname)
    out = Group(name=gname)
    for line in datalines:
        line = line[:-1]
        if line.startswith('#save.'):
            key, value = line[6:].split(':', 1)
            value = value.strip()
            if key == 'nitems': value = int(value)
            header[key] = value
        elif line.startswith('#=>'):
            name = fix_varname(line[4:].strip())
            ivar += 1
            if name in (None, 'None', '__unknown__') or name in varnames:
                name = 'var_%5.5i' % (ivar)
            varnames.append(name)
        else:
            val = decode4js(json.loads(line), grouplist)
            setattr(out, varnames[-1], val)
    setattr(out, '_restore_metadata_', header)

    if top_level:
        _main = _larch.symtable
        for objname in dir(out):
            setattr(_main, objname, getattr(out, objname))
        return
    return out



#!/usr/bin/python

import os
import copy
import numpy as np
from scipy.interpolate import UnivariateSpline

from larch import Group
from larch.xrf import MCA, ROI

def str2floats(s, delim='&'):
    s = s.replace('&', ' ')
    return [float(i) for i in s.split()]

def str2ints(s, delim='&'):
    return [int(i) for i in str2floats(s, delim=delim)]

def str2str(s, delim='&'):
    s = s.strip()
    return [i.strip() for i in s.split(delim) if len(i) > 0]

class GSEMCA_Header(object):
    version  = 'unknown'
    date = ''
    elements = 1
    channels = 2048
    rois = []
    live_time = []
    real_time = []
    cal_slope = []
    cal_offset = []
    cal_quad = []

class GSEMCA_File(Group):
    """
    Read GSECARS style MCA / Multi-element MCA files
    """
    def __init__(self, filename=None, text=None, bad=None, **kws):
        kwargs = {'name': 'GSE MCA File: %s' % filename}
        kwargs.update(kws)
        Group.__init__(self,  **kwargs)
        self.mcas   = []
        self.__mca0 = None
        self.bad    = bad
        if bad is None:
            self.bad = []

        self.filename = filename
        if filename:
            self.read(filename=filename)
        elif text is not None:
            self.readtext(text)

    def __get_mca0(self, chan_min=2, min_counts=2):
        """ find first good detector for alignment
        'good' is defined as at least min_counts counts
        above channel chan_min
        """
        if self.__mca0 is None:
            for imca, mca in enumerate(self.mcas):
                if mca.counts[chan_min:].sum() > min_counts:
                    self.__mca0 = mca
                    self.offset = mca.offset
                    self.slope  = mca.slope
                    self.quad   = mca.quad
                    break
                elif imca not in self.bad:
                    self.bad.append(imca)
            if self.__mca0 is None:
                self.__mca0 = mca = self.mcas[0]
                self.offset = mca.offset
                self.slope  = mca.slope
                self.quad   = mca.quad
        return self.__mca0

    def get_energy(self, imca=None):
        "get energy, optionally selecting which mca to use"
        if imca is not None:
            mca = self.mcas[imca]
        else:
            mca = self.__get_mca0()
        return mca.get_energy()

    def get_counts(self, dt_correct=True, align=True):
        """ get summed MCA spectra,

        Options:
        --------
          align   align spectra in energy before summing (True).
        """
        mca0 = self.__get_mca0()
        en  = mca0.get_energy()
        dat = 0
        for mca in self.mcas:
            mdat = mca.counts
            if align and mca != mca0:
                _en  = mca.get_energy()
                mdat = UnivariateSpline(_en, mdat, s=0)(en)
            if dt_correct:
                mdat = mdat * mca.dt_factor
            dat = dat + mdat
        return dat.astype(np.int)

    def predict_pileup(self, scale=None):
        """predict pileup for an MCA spectra from its auto-correlation.

        Options:
        --------
          scale   factor to apply to convolution [found from data]

        the output `pileup` will be the average of the `pileup` for each MCA

        """
        pileup = self.mcas[0].counts * 0.0
        for m in self.mcas:
            m.predict_pileup(scale=scale)
            pileup += m.pileup
        self.pileup = pileup / len(self.mcas)

    def predict_escape(self, det='Si', scale=1):
        """predict detector escape, save to `escape` attribute

        Options:
        --------
          det       detector material ['Si']
          scale     scale factor [1]

        Outputs 'escape' attribute will contain the average `escape` for each MCA
        """
        escape = self.mcas[0].counts * 0.0
        for m in self.mcas:
            m.predict_escape(det=det, scale=scale)
            escape += m.escape
        self.escape = escape  / len(self.mcas)

    def read(self, filename=None, bad=None):
        """read GSE MCA file"""
        self.filename = filename
        with open(filename, 'r') as fh:
            return self.readtext(fh.read(), bad=bad)

    def readtext(self, text, bad=None):
        """read text of GSE MCA file"""
        lines = text.split('\n')
        if bad is None:
            bad = self.bad
        nrow       = 0
        data_mode  = 'HEADER'
        counts     = []
        rois       = []
        environ    = []
        self.incident_energy = None
        head = self.header = GSEMCA_Header()
        for l in lines:
            l  = l.strip()
            if len(l) < 1: continue
            if data_mode == 'DATA':
                counts.append(str2ints(l))
            else:
                pos = l.find(' ')
                if (pos == -1): pos = len(l)
                tag = l[0:pos].strip().lower()
                if tag.endswith(':'):
                    tag = tag[:-1]
                val = l[pos:len(l)].strip()
                if tag in ('version', 'date'):
                    setattr(head, tag, val)
                elif tag in ('elements', 'channels'):
                    setattr(head, tag, int(val))
                elif tag in ('real_time', 'live_time', 'cal_offset',
                             'cal_slope', 'cal_quad'):
                    setattr(head, tag, str2floats(val))
                elif tag == 'rois':
                    head.rois = str2ints(val)
                    self.nrois = max(head.rois)
                elif tag == 'data':
                    data_mode = 'DATA'
                elif tag == 'environment':
                    addr, val = val.split('="')
                    val, desc = val.split('"')
                    val = val.strip()
                    desc = desc.strip()
                    if desc.startswith('(') and desc.endswith(')'):
                        desc = desc[1:-1]
                    environ.append((desc, val, addr))
                    if 'mono' in desc.lower() and 'energy' in desc.lower():
                        try:
                            val = float(val)
                        except ValueError:
                            pass
                        self.incident_energy = val
                elif tag[0:4] == 'roi_':
                    iroi, item = tag[4:].split('_')
                    iroi = int(iroi)
                    if iroi >= len(rois):
                        for ir in range(1  + iroi - len(rois)):
                            rois.append({'label':[], 'right':[], 'left':[]})
                    if item == "label":
                        rois[iroi]['label'] = str2str(val, delim='&')
                    elif item == "left":
                        rois[iroi]['left']  = str2ints(val)
                    elif item == "right":
                        rois[iroi]['right'] = str2ints(val)
                else:
                    pass # print(" Warning: " , tag, " is not supported here!")

        #
        counts =  np.array(counts)
        ## Data has been read, now store in MCA objects
        sum_mca = None

        for tag in ('real_time', 'live_time', 'cal_offset',
                    'cal_slope', 'cal_quad'):
            val = getattr(head, tag)
            # print( ' Attr ', tag, val)
            if len(val) == 1 and head.elements > 1:
                val = [val[0]]*head.elements
                setattr(head, tag, val)
        for imca in range(head.elements):
            thismca = MCA(name='mca%i' % (imca+1),
                          nchans=head.channels,
                          counts=counts[:,imca],
                          start_time=head.date,
                          offset=head.cal_offset[imca],
                          slope=head.cal_slope[imca],
                          quad=head.cal_quad[imca],
                          real_time=head.real_time[imca],
                          live_time=head.live_time[imca])

            for desc, val, addr in environ:
                thismca.add_environ(desc=desc, val=val, addr=addr)

            for roi in rois:
                left = roi['left'][imca]
                right = roi['right'][imca]
                label = roi['label'][imca]
                if right > 1 and len(label) > 1:
                    thismca.add_roi(name=label, left=left, right=right,
                                    sort=False, counts=counts[:,imca])
            thismca.rois.sort()
            self.mcas.append(thismca)

        mca0 = self.__get_mca0()
        self.counts = self.get_counts()
        self.raw    = self.get_counts(dt_correct=False)
        self.name   = 'mcasum'
        self.energy = mca0.energy[:]
        self.environ = mca0.environ
        self.real_time = mca0.real_time
        self.live_time = mca0.live_time
        self.offset = mca0.offset
        self.slope  = mca0.slope
        self.quad   = mca0.quad
        self.rois = []
        for roi in mca0.rois:
            self.add_roi(name=roi.name, left=roi.left,
                         right=roi.right, sort=False,
                         counts=counts, to_mcas=False)
        self.rois.sort()
        return

    def add_roi(self, name='', left=0, right=0, bgr_width=3,
                counts=None, sort=True, to_mcas=True):
        """add an ROI to the sum spectra"""
        name = name.strip()
        # print('GSEMCA: Add ROI ', name, left, right)
        roi = ROI(name=name, left=left, right=right,
                  bgr_width=bgr_width, counts=counts)
        rnames = [r.name.lower() for r in self.rois]
        if name.lower() in rnames:
            iroi = rnames.index(name.lower())
            self.rois[iroi] = roi
        else:
            self.rois.append(roi)
        if sort:
            self.rois.sort()
        if to_mcas:
            mca0 = self.__get_mca0()
            slo0 = mca0.slope
            off0 = mca0.offset
            mca0.add_roi(name=name, left=left, right=right,
                         bgr_width=bgr_width)
            for mca in self.mcas:
                if mca != mca0:
                    xleft  = int(0.5 + ((off0 + left*slo0) - mca.offset)/mca.slope)
                    xright = int(0.5 + ((off0 + right*slo0) - mca.offset)/mca.slope)
                    mca.add_roi(name=name, left=xleft, right=xright,
                                 bgr_width=bgr_width)

    def save_mcafile(self, filename):
        """
        write multi-element MCA file
        Parameters:
        -----------
        * filename: output file name
        """
        with open(filename, 'w') as fh:
            fh.write(self.dump_mcafile())

    def dump_mcafile(self):
        """return text of MCA file, not writing to disk, as for dumping"""
        nchans = len(self.counts)
        ndet   = len(self.mcas)

        # formatted count times and calibration
        rtimes  = ["%f" % m.real_time for m in self.mcas]
        ltimes  = ["%f" % m.live_time for m in self.mcas]
        offsets = ["%e" % m.offset    for m in self.mcas]
        slopes  = ["%e" % m.slope     for m in self.mcas]
        quads   = ["%e" % m.quad      for m in self.mcas]

        b = ['VERSION:    3.1']
        b.append('ELEMENTS:   %i' % ndet)
        b.append('DATE:       %s' % self.mcas[0].start_time)
        b.append('CHANNELS:   %i' % nchans)
        b.append('REAL_TIME:  %s' % ' '.join(rtimes))
        b.append('LIVE_TIME:  %s' % ' '.join(ltimes))
        b.append('CAL_OFFSET: %s' % ' '.join(offsets))
        b.append('CAL_SLOPE:  %s' % ' '.join(slopes))
        b.append('CAL_QUAD:   %s' % ' '.join(quads))

        # Write ROIS  in channel units
        nrois = ["%i" % len(m.rois) for m in self.mcas]
        rois = [m.rois for m in self.mcas]
        b.append('ROIS:      %s' % ' '.join(nrois))

        # don't assume number of ROIS is same for all elements
        nrois = max([len(r) for r in rois])
        # print('NROIS ' , nrois, [len(r) for r in rois])
        for ir, r in enumerate(rois):
            if len(r) < nrois:
                for i in range(nrois - len(r)):
                    r.append(ROI(name='', left=0, right=0))
        # print( 'NROIS ' , nrois, [len(r) for r in rois])
        for i in range(len(rois[0])):
            names = ' &  '.join([r[i].name  for r in rois])
            left  = ' '.join(['%i' % r[i].left  for r in rois])
            right = ' '.join(['%i' % r[i].right for r in rois])
            b.append('ROI_%i_LEFT:  %s' % (i, left))
            b.append('ROI_%i_RIGHT:  %s' % (i, right))
            b.append('ROI_%i_LABEL: %s &' % (i, names))

        # environment
        for e in self.environ:
            b.append('ENVIRONMENT: %s="%s" (%s)' % (e.addr, e.val, e.desc))
        # data
        b.append('DATA: ')
        for i in range(nchans):
            d = ' '.join(["%i" % m.counts[i] for m in self.mcas])
            b.append(" %s" % d)
        b.append('')
        return '\n'.join(b)

    def save_ascii(self, filename):
        """
        write multi-element MCA file to XDI-style ASCII file
        Parameters:
        -----------
        * filename: output file name
        """
        nchans = len(self.counts)
        ndet   = len(self.mcas)


        mca0 = self.mcas[0]
        buff = ['# XDI/1.0  GSE/1.0',
                '# Collection.date: %s' % mca0.start_time,
                '# Collection.n_detectors: %i' % ndet,
                '# Collection.n_channels: %i' % nchans,
                '# Collection.real_time: %i' % mca0.real_time,
                '# Collection.live_time: %s' % mca0.live_time,
                '# Calibration.offset: %s' % mca0.offset,
                '# Calibration.slope: %s' % mca0.slope,
                '# Calibration.quad: %s' % mca0.quad,
                '# Column.1: energy keV']

        label = '#   energy  '
        for i in range(ndet):
            buff.append('# Column.%i: MCA%i counts' % (i+2, i+1))
            label = '%s    MCA%i '  % (label, i+1)

        froi = '# ROI.%i: %s [%i:%i]'
        fenv = '# ENV.%s:  %s [%s]'
        for i, roi in enumerate(mca0.rois):
            buff.append(froi % (i, roi.name, roi.left, roi.right))


        for e in self.environ:
            desc = e.desc.replace(' ', '_')
            buff.append(fenv % (desc, e.val, e.addr))
        buff.append('#--------------------')
        buff.append(label)

        # data
        for i in range(nchans):
            d = ['%9.3f' % self.energy[i]]
            d.extend(['%11i' % m.counts[i] for m in self.mcas])
            buff.append(' %s' % ' '.join(d))
        buff.append('')

        fp = open(filename, 'w')
        fp.write('\n'.join(buff))
        fp.close()


def gsemca_group(filename=None, text=None, _larch=None, **kws):
    """read GSECARS MCA file to larch group"""
    return GSEMCA_File(filename=filename, text=text)



import os
import ctypes
import numpy as np
from scipy.signal import convolve

import larch
from larch.larchlib import get_dll

from larch.math import as_ndarray
from xraydb import core_width, atomic_number

CLLIB = None

def f1f2(z, energies, width=None, edge=None):
    """Return anomalous scattering factors f1, f2 from Cromer-Liberman

    Look-up and return f1, f2 for an element and array of energies
    from Cromer-Liberman (Cowan-Brennan implementation)

    Parameters
    ----------
    z:         atomic number of element
    energies:  array of x-ray energies (in eV)
    width:     width used to convolve values with lorentzian profile
    edge:      x-ray edge ('K', 'L3', etc) used to lookup energy
               width for convolution.

    Returns:
    ---------
    f1, f2:    anomalous scattering factors

    """
    global CLLIB
    if CLLIB is None:
        CLLIB = get_dll('cldata')

    en = as_ndarray(energies)

    if not isinstance(z, int):
        z  = atomic_number(z)
        if z is None:
            return None

    if z > 92:
        print( 'Cromer-Liberman data not available for Z>92')
        return

    if edge is not None or width is not None:
        natwid = core_width(element=z, edge=edge)
        if width is None and natwid not in (None, []):
            width = natwid

    if width is not None: # will convolve!
        e_extra = int(width*80.0)
        estep = (en[1:] - en[:-1]).min()
        emin = min(en) - e_extra
        emax = max(en) + e_extra

        npts = 1 + abs(emax-emin+estep*0.02)/abs(estep)
        en   = np.linspace(emin, emax, int(npts))
        nk   = int(e_extra / estep)
        sig  = width/2.0
        lor  = (1./(1 + ((np.arange(2*nk+1)-nk*1.0)/sig)**2))/(np.pi*sig)
        scale = lor.sum()

    # create ctypes pointers for the C function
    npts   = len(en)
    p_z    = ctypes.pointer(ctypes.c_int(int(z)))
    p_npts = ctypes.pointer(ctypes.c_int(npts))
    p_en   = (npts*ctypes.c_double)()
    p_f1   = (npts*ctypes.c_double)()
    p_f2   = (npts*ctypes.c_double)()

    for i in range(npts):
        p_en[i] = en[i]

    nout = CLLIB.f1f2(p_z, p_npts, p_en, p_f1, p_f2)
    f1 = np.array([i for i in p_f1[:]])
    f2 = np.array([i for i in p_f2[:]])
    if width is not None: # do the convolution
        f1 = np.interp(energies, en, convolve(f1, lor)[nk:-nk])/scale
        f2 = np.interp(energies, en, convolve(f2, lor)[nk:-nk])/scale
    return (f1, f2)

if __name__ == '__main__':
    en = np.linspace(8000, 9200, 51)
    f1, f2 = f1f2(29, en)
    print( en, f1, f2)
